from pathlib import Path

from rak.registry import load_registry, save_registry, compute_hash


def test_save_and_load_roundtrip(tmp_path: Path):
    registry = {"ABC123": "a1b2c3", "DEF456": "d4e5f6"}
    save_registry(tmp_path, registry)
    loaded = load_registry(tmp_path)
    assert loaded == registry


def test_load_missing_returns_empty_dict(tmp_path: Path):
    assert load_registry(tmp_path) == {}


def test_compute_hash_deterministic():
    text = "Single Cell Analysis\nAuthors: John Doe"
    assert compute_hash(text) == compute_hash(text)


def test_compute_hash_different_texts():
    assert compute_hash("hello") != compute_hash("world")


def test_load_corrupt_json_returns_empty(tmp_path: Path):
    (tmp_path / "registry.json").write_text("{corrupt json!!!")
    assert load_registry(tmp_path) == {}


def test_load_non_dict_returns_empty(tmp_path: Path):
    """If registry.json contains a list instead of dict, return empty."""
    import json
    (tmp_path / "registry.json").write_text(json.dumps(["a", "b"]))
    assert load_registry(tmp_path) == {}


def test_load_coerces_keys_and_values_to_strings(tmp_path: Path):
    import json
    (tmp_path / "registry.json").write_text(json.dumps({123: 456}))
    result = load_registry(tmp_path)
    assert result == {"123": "456"}


def test_save_registry_atomic_write(tmp_path: Path):
    """save_registry uses atomic write — no leftover tmp files on success."""
    registry = {"K1": "hash1", "K2": "hash2"}
    save_registry(tmp_path, registry)
    target = tmp_path / "registry.json"
    assert target.exists()
    import json
    assert json.loads(target.read_text()) == registry
    # No leftover .tmp files
    tmp_files = list(tmp_path.glob("*.tmp"))
    assert tmp_files == []


def test_save_registry_overwrites_existing(tmp_path: Path):
    import json
    save_registry(tmp_path, {"A": "1"})
    save_registry(tmp_path, {"B": "2"})
    loaded = load_registry(tmp_path)
    assert loaded == {"B": "2"}



from unittest.mock import patch, MagicMock

from rak.mcp_server import (
    mcp, search_papers, search_papers_bm25, similar_papers,
    ask_papers, export_papers, show_config, index_status,
    _resolve_key_mcp,
)


def test_mcp_server_loads():
    assert mcp is not None
    assert mcp.name == "rak"


def test_index_status_no_index(tmp_path):
    from rak.config import RakConfig
    fake_config = RakConfig(data_dir=tmp_path)
    with patch("rak.mcp_server._get_config", return_value=fake_config):
        result = index_status()
    assert "No index found" in result


def test_index_status_with_index(tmp_path):
    import json
    from rak.config import RakConfig
    from rak.metadata import save_metadata
    fake_config = RakConfig(data_dir=tmp_path)
    save_metadata(tmp_path, "all-MiniLM-L6-v2", 42)
    with patch("rak.mcp_server._get_config", return_value=fake_config):
        result = index_status()
    data = json.loads(result)
    assert data["item_count"] == 42
    assert data["model_name"] == "all-MiniLM-L6-v2"


def test_search_papers_returns_json():
    import json
    from rak.searcher import SearchResult

    mock_searcher = MagicMock()
    mock_searcher.vector_search.return_value = [
        SearchResult(doc_id="A1", score=0.9, title="Paper One", source="vector"),
    ]
    mock_bm25 = MagicMock()

    with patch("rak.mcp_server._get_config"), \
         patch("rak.mcp_server._init_searcher", return_value=(mock_searcher, None, mock_bm25)):
        result = search_papers("test query", limit=5)

    data = json.loads(result)
    assert len(data) == 1
    assert data[0]["key"] == "A1"
    assert data[0]["title"] == "Paper One"
    assert "text" not in data[0]


def test_search_papers_bm25_returns_json(tmp_path):
    import json
    from rak.config import RakConfig
    from rak.searcher import SearchResult

    fake_config = RakConfig(data_dir=tmp_path)

    mock_searcher = MagicMock()
    mock_searcher.bm25_search.return_value = [
        SearchResult(doc_id="B1", score=5.0, title="Keyword Paper", source="bm25", snippet="matched text"),
    ]

    with patch("rak.mcp_server._get_config", return_value=fake_config), \
         patch("rak.bm25.BM25Index.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.close"), \
         patch("rak.searcher.Searcher.__init__", return_value=None), \
         patch("rak.searcher.Searcher.bm25_search", return_value=mock_searcher.bm25_search.return_value):
        result = search_papers_bm25("keyword query", limit=5)

    data = json.loads(result)
    assert len(data) == 1
    assert data[0]["key"] == "B1"
    assert data[0]["source"] == "bm25"
    assert data[0]["snippet"] == "matched text"


def test_similar_papers_returns_json():
    import json
    from rak.searcher import SearchResult

    mock_searcher = MagicMock()
    mock_searcher.similar_search.return_value = [
        SearchResult(doc_id="C1", score=0.85, title="Similar Paper", source="similar"),
    ]
    mock_bm25 = MagicMock()

    with patch("rak.mcp_server._get_config"), \
         patch("rak.mcp_server._init_searcher", return_value=(mock_searcher, None, mock_bm25)), \
         patch("rak.mcp_server._resolve_key_mcp", return_value=("KEY1", None)):
        result = similar_papers("KEY1", limit=5)

    data = json.loads(result)
    assert len(data) == 1
    assert data[0]["key"] == "C1"
    assert data[0]["source"] == "similar"
    mock_searcher.similar_search.assert_called_once_with("KEY1", limit=5, collection=None, tags=None)


def test_similar_papers_title_search():
    import json

    mock_searcher = MagicMock()
    mock_bm25 = MagicMock()
    candidates = [{"key": "A1", "title": "Paper One"}, {"key": "A2", "title": "Paper Two"}]

    with patch("rak.mcp_server._get_config"), \
         patch("rak.mcp_server._init_searcher", return_value=(mock_searcher, None, mock_bm25)), \
         patch("rak.mcp_server._resolve_key_mcp", return_value=(None, candidates)):
        result = similar_papers("attention", limit=5)

    data = json.loads(result)
    assert data["status"] == "multiple_matches"
    assert len(data["candidates"]) == 2


def test_ask_papers_returns_answer():
    import json
    from rak.searcher import SearchResult

    mock_searcher = MagicMock()
    mock_searcher.vector_search.return_value = [
        SearchResult(doc_id="D1", score=0.9, title="Context Paper", source="vector", snippet="some context"),
    ]
    mock_bm25 = MagicMock()
    mock_llm = MagicMock()
    mock_llm.ask.return_value = "The answer is 42."

    with patch("rak.mcp_server._get_config"), \
         patch("rak.mcp_server._init_searcher", return_value=(mock_searcher, None, mock_bm25)), \
         patch("rak.llm.LLMClient", return_value=mock_llm):
        result = ask_papers("What is the answer?")

    data = json.loads(result)
    assert data["answer"] == "The answer is 42."
    assert len(data["sources"]) == 1
    assert data["sources"][0]["key"] == "D1"


def test_ask_papers_no_results():
    import json

    mock_searcher = MagicMock()
    mock_searcher.vector_search.return_value = []
    mock_bm25 = MagicMock()

    with patch("rak.mcp_server._get_config"), \
         patch("rak.mcp_server._init_searcher", return_value=(mock_searcher, None, mock_bm25)):
        result = ask_papers("unanswerable question")

    data = json.loads(result)
    assert "No relevant papers" in data["answer"]
    assert data["sources"] == []


def test_export_papers_csv():
    from rak.searcher import SearchResult

    mock_searcher = MagicMock()
    mock_searcher.vector_search.return_value = [
        SearchResult(doc_id="E1", score=0.8, title="Export Paper", source="vector"),
    ]
    mock_vs = MagicMock()
    mock_vs.get.return_value = {"ids": ["E1"], "metadatas": [{"date": "2024", "item_type": "article"}]}
    mock_bm25 = MagicMock()

    with patch("rak.mcp_server._get_config"), \
         patch("rak.mcp_server._init_searcher", return_value=(mock_searcher, mock_vs, mock_bm25)):
        result = export_papers("test query")

    assert "key,title,score,source" in result
    assert "E1" in result


def test_export_papers_bibtex():
    from rak.searcher import SearchResult

    mock_searcher = MagicMock()
    mock_searcher.vector_search.return_value = [
        SearchResult(doc_id="F1", score=0.7, title="BibTeX Paper", source="vector"),
    ]
    mock_vs = MagicMock()
    mock_vs.get.return_value = {"ids": ["F1"], "metadatas": [{"date": "2024", "item_type": "conferencePaper"}]}
    mock_bm25 = MagicMock()

    with patch("rak.mcp_server._get_config"), \
         patch("rak.mcp_server._init_searcher", return_value=(mock_searcher, mock_vs, mock_bm25)):
        result = export_papers("test query", format="bibtex")

    assert "@inproceedings{F1," in result


def test_show_config(tmp_path):
    import json
    from rak.config import RakConfig

    fake_config = RakConfig(data_dir=tmp_path)
    with patch("rak.mcp_server._get_config", return_value=fake_config):
        result = show_config()

    data = json.loads(result)
    assert data["model_name"] == "all-MiniLM-L6-v2"
    assert data["pdf_provider"] == "pymupdf"
    assert data["chunk_size"] == 512


def test_cache_invalidation_on_missing_db():
    """If FTS db is deleted, cached searcher should be invalidated."""
    import rak.mcp_server as mcp_mod

    mock_searcher_old = MagicMock()
    mock_vs_old = MagicMock()
    mock_bm25_old = MagicMock()
    # Simulate probe failing (db gone)
    mock_bm25_old.search.side_effect = Exception("db gone")

    mock_searcher_new = MagicMock()
    mock_vs_new = MagicMock()
    mock_bm25_new = MagicMock()

    old_cached = mcp_mod._cached_searcher
    try:
        mcp_mod._cached_searcher = (mock_searcher_old, mock_vs_old, mock_bm25_old)

        fake_config = MagicMock()
        fake_config.fts_db_path.exists.return_value = False  # db missing

        with patch("rak.bm25.BM25Index", return_value=mock_bm25_new), \
             patch("rak.embedder.Embedder", return_value=MagicMock(dimension=384)), \
             patch("rak.store.VectorStore", return_value=mock_vs_new), \
             patch("rak.searcher.Searcher", return_value=mock_searcher_new):
            result_tuple = mcp_mod._init_searcher(fake_config)

        # Old bm25 should have been closed
        mock_bm25_old.close.assert_called_once()
        # Cache should have been rebuilt — new searcher returned
        searcher, vs, bm25 = result_tuple
        assert searcher is mock_searcher_new
    finally:
        mcp_mod._cached_searcher = old_cached


def test_cleanup_function():
    """Test _cleanup closes BM25 and clears cache."""
    import rak.mcp_server as mcp_mod
    from rak.mcp_server import _cleanup

    mock_bm25 = MagicMock()
    old_cached = mcp_mod._cached_searcher
    try:
        mcp_mod._cached_searcher = (MagicMock(), MagicMock(), mock_bm25)
        _cleanup()
        mock_bm25.close.assert_called_once()
        assert mcp_mod._cached_searcher is None
    finally:
        mcp_mod._cached_searcher = old_cached


def test_cleanup_when_no_cache():
    """_cleanup should be safe to call when nothing is cached."""
    import rak.mcp_server as mcp_mod
    from rak.mcp_server import _cleanup

    old_cached = mcp_mod._cached_searcher
    try:
        mcp_mod._cached_searcher = None
        _cleanup()  # should not raise
        assert mcp_mod._cached_searcher is None
    finally:
        mcp_mod._cached_searcher = old_cached


def test_export_papers_no_results():
    mock_searcher = MagicMock()
    mock_searcher.vector_search.return_value = []
    mock_bm25 = MagicMock()

    with patch("rak.mcp_server._get_config"), \
         patch("rak.mcp_server._init_searcher", return_value=(mock_searcher, None, mock_bm25)):
        result = export_papers("nonexistent")

    assert "No results found" in result


def test_similar_papers_not_found():
    import json

    mock_searcher = MagicMock()
    mock_bm25 = MagicMock()

    with patch("rak.mcp_server._get_config"), \
         patch("rak.mcp_server._init_searcher", return_value=(mock_searcher, None, mock_bm25)), \
         patch("rak.mcp_server._resolve_key_mcp", return_value=(None, None)):
        result = similar_papers("nonexistent")

    data = json.loads(result)
    assert data["status"] == "not_found"


def test_search_papers_hybrid():
    import json
    from rak.searcher import SearchResult

    mock_searcher = MagicMock()
    mock_searcher.hybrid_search.return_value = [
        SearchResult(doc_id="H1", score=0.8, title="Hybrid", source="fused"),
    ]
    mock_bm25 = MagicMock()

    with patch("rak.mcp_server._get_config"), \
         patch("rak.mcp_server._init_searcher", return_value=(mock_searcher, None, mock_bm25)):
        result = search_papers("query", hybrid=True)

    data = json.loads(result)
    assert data[0]["source"] == "fused"
    mock_searcher.hybrid_search.assert_called_once()



from __future__ import annotations

import logging
import subprocess
import tempfile
from pathlib import Path

logger = logging.getLogger(__name__)


def _extract_via_pymupdf(pdf_path: Path) -> str:
    """Extract text from a PDF using PyMuPDF (fitz)."""
    try:
        import fitz
        doc = fitz.open(str(pdf_path))
        text = "\n".join(page.get_text() for page in doc)
        doc.close()
        return text.strip()
    except Exception as exc:
        logger.warning("Failed to extract text from %s: %s", pdf_path, exc)
        return ""


def _extract_via_mineru(pdf_path: Path) -> str | None:
    """Call MinerU CLI to parse a PDF and return the Markdown text.
    Returns None on any failure so the caller can fall back.
    """
    with tempfile.TemporaryDirectory() as tmp_dir:
        try:
            result = subprocess.run(
                ["mineru", "-p", str(pdf_path), "-o", tmp_dir],
                capture_output=True, text=True, encoding="utf-8", timeout=300,
            )
            if result.returncode != 0:
                logger.warning("MinerU failed for %s: %s", pdf_path, result.stderr)
                return None
            stem = pdf_path.stem
            md_file = Path(tmp_dir) / stem / "auto" / f"{stem}.md"
            if md_file.exists():
                return md_file.read_text(encoding="utf-8").strip()
            md_files = list(Path(tmp_dir).rglob("*.md"))
            if md_files:
                return md_files[0].read_text(encoding="utf-8").strip()
            logger.warning("MinerU produced no markdown for %s", pdf_path)
            return None
        except Exception as exc:
            logger.warning("MinerU error for %s: %s", pdf_path, exc)
            return None


def _extract_via_docling(pdf_path: Path) -> str | None:
    """Call Docling CLI to parse a PDF and return the Markdown text.
    Returns None on any failure so the caller can fall back.
    """
    with tempfile.TemporaryDirectory() as tmp_dir:
        try:
            result = subprocess.run(
                ["docling", str(pdf_path), "--output", tmp_dir, "--format", "md"],
                capture_output=True, text=True, encoding="utf-8", timeout=300,
            )
            if result.returncode != 0:
                logger.warning("Docling failed for %s: %s", pdf_path, result.stderr)
                return None
            md_files = list(Path(tmp_dir).rglob("*.md"))
            if md_files:
                return md_files[0].read_text(encoding="utf-8").strip()
            logger.warning("Docling produced no markdown for %s", pdf_path)
            return None
        except Exception as exc:
            logger.warning("Docling error for %s: %s", pdf_path, exc)
            return None


def extract_pdf_text(pdf_path: Path, provider: str = "pymupdf") -> str:
    if not pdf_path.exists():
        return ""
    if provider in ("mineru", "docling"):
        extractor = _extract_via_mineru if provider == "mineru" else _extract_via_docling
        text = extractor(pdf_path)
        if text is not None:
            return text
        logger.warning("Falling back to PyMuPDF for %s", pdf_path)
    return _extract_via_pymupdf(pdf_path)


def extract_file_text(file_path: Path, provider: str = "pymupdf") -> str:
    """Extract text from a PDF or Markdown file."""
    suffix = file_path.suffix.lower()
    if suffix == ".pdf":
        return extract_pdf_text(file_path, provider=provider)
    if suffix == ".md":
        try:
            return file_path.read_text(encoding="utf-8").strip()
        except Exception as exc:
            logger.warning("Failed to read %s: %s", file_path, exc)
            return ""
    return ""


def find_attachments(storage_dir: Path, item_key: str) -> list[Path]:
    """Find all PDF and Markdown files in a Zotero item's storage directory."""
    key_dir = storage_dir / item_key
    if not key_dir.is_dir():
        return []
    files = list(key_dir.glob("*.pdf")) + list(key_dir.glob("*.md"))
    return sorted(files)


def _split_paragraphs(text: str) -> list[str]:
    """Split text into paragraphs on double newlines and markdown headings."""
    import re
    # Split on double newlines or before markdown headings
    parts = re.split(r'\n\s*\n|(?=\n#{1,6}\s)', text)
    return [p.strip() for p in parts if p.strip()]


def _chunk_words(text: str, chunk_size: int, overlap: int) -> list[str]:
    """Fixed-size word-level chunking with overlap (fallback for oversized paragraphs)."""
    words = text.split()
    if not words:
        return []
    chunks = []
    start = 0
    step = max(chunk_size - overlap, 1)
    while start < len(words):
        end = start + chunk_size
        chunks.append(" ".join(words[start:end]))
        start += step
    return chunks


def chunk_text(
    text: str,
    chunk_size: int = 512,
    overlap: int = 64,
) -> list[str]:
    """Split text into chunks, preferring paragraph/section boundaries.

    Splits on double newlines and markdown headings first, then merges
    small consecutive paragraphs up to chunk_size words. Oversized
    paragraphs are split with word-level overlap. Returns a single-element
    list if the text fits in one chunk.
    """
    if overlap >= chunk_size:
        raise ValueError(f"chunk_overlap ({overlap}) must be less than chunk_size ({chunk_size})")
    words = text.split()
    if not words:
        return []
    if len(words) <= chunk_size:
        return [text]

    paragraphs = _split_paragraphs(text)
    if len(paragraphs) <= 1:
        # No paragraph structure found, fall back to word-level chunking
        return _chunk_words(text, chunk_size, overlap)

    # Merge small paragraphs, split oversized ones
    chunks: list[str] = []
    current_parts: list[str] = []
    current_words = 0

    for para in paragraphs:
        para_words = len(para.split())
        if para_words > chunk_size:
            # Flush accumulated parts first
            if current_parts:
                chunks.append("\n\n".join(current_parts))
                current_parts = []
                current_words = 0
            # Split oversized paragraph with overlap
            chunks.extend(_chunk_words(para, chunk_size, overlap))
        elif current_words + para_words > chunk_size and current_parts:
            # Current buffer is full, flush it
            chunks.append("\n\n".join(current_parts))
            current_parts = [para]
            current_words = para_words
        else:
            current_parts.append(para)
            current_words += para_words

    if current_parts:
        chunks.append("\n\n".join(current_parts))

    return chunks



from unittest.mock import MagicMock, patch

from rak.chat import ChatSession, estimate_tokens, HELP_TEXT
from rak.searcher import SearchResult


def _mock_searcher(results):
    searcher = MagicMock()
    searcher.vector_search.return_value = results
    searcher.hybrid_search.return_value = results
    return searcher


def _mock_llm(tokens):
    llm = MagicMock()
    llm.stream_messages.return_value = iter(tokens)
    return llm


def test_chat_session_search_populates_context():
    results = [
        SearchResult(doc_id="A1", score=0.9, title="Paper One", source="vector",
                     snippet="Full text of paper one."),
    ]
    searcher = _mock_searcher(results)
    llm = _mock_llm([])

    session = ChatSession(searcher=searcher, llm=llm)
    session.search("test query")

    assert len(session.context) == 1
    assert session.context[0]["key"] == "A1"
    assert session.context[0]["text"] == "Full text of paper one."
    assert len(session.messages) == 1  # system message only
    assert session.messages[0]["role"] == "system"


def test_chat_session_ask_appends_history():
    results = [
        SearchResult(doc_id="A1", score=0.9, title="Paper One", source="vector",
                     snippet="Text."),
    ]
    searcher = _mock_searcher(results)
    llm = _mock_llm(["The ", "answer."])

    session = ChatSession(searcher=searcher, llm=llm)
    session.search("query")

    tokens = list(session.ask("What is this about?"))
    assert tokens == ["The ", "answer."]
    assert len(session.messages) == 3  # system + user + assistant
    assert session.messages[1]["role"] == "user"
    assert session.messages[1]["content"] == "What is this about?"
    assert session.messages[2]["role"] == "assistant"
    assert session.messages[2]["content"] == "The answer."


def test_chat_session_search_resets_history():
    results = [
        SearchResult(doc_id="A1", score=0.9, title="Paper One", source="vector",
                     snippet="Text."),
    ]
    searcher = _mock_searcher(results)
    llm = _mock_llm(["Response."])

    session = ChatSession(searcher=searcher, llm=llm)

    session.search("first query")
    list(session.ask("question"))  # adds user + assistant messages

    assert len(session.messages) == 3

    # Re-search should reset
    llm.stream_messages.return_value = iter(["New."])
    session.search("second query")
    assert len(session.messages) == 1  # only system message
    assert session.messages[0]["role"] == "system"


def test_estimate_tokens():
    assert estimate_tokens("hello world") > 0
    assert estimate_tokens("") == 0
    # ~4 chars per token
    assert estimate_tokens("a" * 400) == 100


def test_token_count_property():
    results = [SearchResult(doc_id="A1", score=0.9, title="P", source="vector", snippet="Text.")]
    searcher = _mock_searcher(results)
    llm = _mock_llm(["Reply."])
    session = ChatSession(searcher=searcher, llm=llm)
    session.search("query")
    initial_tokens = session.token_count
    assert initial_tokens > 0
    list(session.ask("question"))
    assert session.token_count > initial_tokens


def test_turn_count_property():
    results = [SearchResult(doc_id="A1", score=0.9, title="P", source="vector", snippet="T.")]
    searcher = _mock_searcher(results)
    llm = _mock_llm(["R."])
    session = ChatSession(searcher=searcher, llm=llm)
    session.search("q")
    assert session.turn_count == 0
    list(session.ask("q1"))
    assert session.turn_count == 1


def test_help_text_contains_commands():
    assert "/search" in HELP_TEXT
    assert "/context" in HELP_TEXT
    assert "/tokens" in HELP_TEXT
    assert "/help" in HELP_TEXT
    assert "/quit" in HELP_TEXT


def test_chat_session_bm25_only_mode():
    results = [
        SearchResult(doc_id="K1", score=5.0, title="", source="bm25", snippet="keyword match"),
    ]
    searcher = MagicMock()
    searcher.bm25_search.return_value = results
    llm = _mock_llm(["Answer."])

    session = ChatSession(searcher=searcher, llm=llm, bm25_only=True)
    session.search("keyword query")

    assert len(session.context) == 1
    assert session.context[0]["key"] == "K1"
    searcher.bm25_search.assert_called_once()
    searcher.vector_search.assert_not_called()
    searcher.hybrid_search.assert_not_called()



"""MCP server for rak — exposes search and status tools for LM Studio, Cursor, etc."""
from __future__ import annotations

import atexit
import json
import threading

from mcp.server.fastmcp import FastMCP

from rak.config import RakConfig

mcp = FastMCP("rak")

_cached_searcher: tuple | None = None
_searcher_lock = threading.Lock()


def _cleanup() -> None:
    """Close cached BM25 SQLite connection on shutdown."""
    global _cached_searcher
    with _searcher_lock:
        if _cached_searcher is not None:
            _, _, bm25 = _cached_searcher
            bm25.close()
            _cached_searcher = None


atexit.register(_cleanup)


def _get_config() -> RakConfig:
    return RakConfig()


def _init_searcher(config: RakConfig):
    global _cached_searcher
    with _searcher_lock:
        if _cached_searcher is not None:
            # Validate cached BM25 connection is still usable
            _, _, bm25 = _cached_searcher
            try:
                if not config.fts_db_path.exists():
                    raise FileNotFoundError
                bm25.search("", limit=1)  # lightweight probe
            except Exception:
                bm25.close()
                _cached_searcher = None

        if _cached_searcher is not None:
            return _cached_searcher

        from rak.bm25 import BM25Index
        from rak.embedder import Embedder
        from rak.searcher import Searcher
        from rak.store import VectorStore

        embedder = Embedder(config.model_name, provider=config.embedding_provider, base_url=config.embedding_base_url, api_key=config.embedding_api_key, batch_size=config.embedding_batch_size)
        vector_store = VectorStore(config.chroma_dir, embedder.dimension)
        bm25 = BM25Index(config.fts_db_path)
        searcher = Searcher(embedder, vector_store, bm25)
        _cached_searcher = (searcher, vector_store, bm25)
        return _cached_searcher


@mcp.tool()
def search_papers(query: str, limit: int = 10, hybrid: bool = False,
                  collection: str | None = None, tags: list[str] | None = None) -> str:
    """Search your Zotero library using semantic or hybrid search.

    Args:
        query: Search query (natural language)
        limit: Maximum number of results (default 10)
        hybrid: Use hybrid search combining vector + BM25 keyword search
        collection: Filter by Zotero collection name
        tags: Filter by tags (OR logic)

    Returns:
        JSON array of search results with key, title, score, and source.
    """
    config = _get_config()
    searcher, _, bm25 = _init_searcher(config)

    tag_list = tags if tags else None
    if hybrid:
        results = searcher.hybrid_search(query, limit=limit, collection=collection, tags=tag_list)
    else:
        results = searcher.vector_search(query, limit=limit, collection=collection, tags=tag_list)

    output = []
    for r in results:
        item = {"key": r.doc_id, "title": r.title, "score": round(r.score, 4), "source": r.source}
        if r.snippet:
            item["snippet"] = r.snippet
        output.append(item)

    return json.dumps(output, indent=2, ensure_ascii=False)


@mcp.tool()
def search_papers_bm25(query: str, limit: int = 10) -> str:
    """Search your Zotero library using pure BM25 keyword search (no embedding model needed).

    Args:
        query: Search query (keywords)
        limit: Maximum number of results (default 10)

    Returns:
        JSON array of search results with key, title, score, snippet, and source.
    """
    from rak.bm25 import BM25Index
    from rak.searcher import Searcher

    config = _get_config()
    with BM25Index(config.fts_db_path) as bm25:
        searcher = Searcher(None, None, bm25)
        results = searcher.bm25_search(query, limit=limit)

    output = []
    for r in results:
        item = {"key": r.doc_id, "title": r.title, "score": round(r.score, 4), "source": r.source}
        if r.snippet:
            item["snippet"] = r.snippet
        output.append(item)

    return json.dumps(output, indent=2, ensure_ascii=False)


def _resolve_key_mcp(key_or_title: str, vector_store, bm25) -> tuple[str | None, list[dict] | None]:
    """Resolve a key or title to a Zotero key for MCP (non-interactive).

    Returns (resolved_key, candidates_or_none). If multiple matches,
    returns (None, candidates) so the caller can return them to the AI.
    """
    import re
    if re.match(r'^[A-Za-z0-9]{4,12}$', key_or_title):
        if vector_store.has(key_or_title) or vector_store.has(f"{key_or_title}_chunk_0"):
            return key_or_title, None

    results = bm25.search(key_or_title, limit=10)
    if not results:
        return None, None

    parent_ids = list(dict.fromkeys(
        doc_id.split("_chunk_")[0] if "_chunk_" in doc_id else doc_id
        for doc_id in [r["id"] for r in results]
    ))
    doc_data = vector_store.get(ids=parent_ids, include=["metadatas"])
    candidates = [{"key": doc_id, "title": meta.get("title", doc_id)}
                  for doc_id, meta in zip(doc_data["ids"], doc_data["metadatas"])]

    if len(candidates) == 1:
        return candidates[0]["key"], None
    return None, candidates


@mcp.tool()
def similar_papers(key_or_title: str, limit: int = 10,
                   collection: str | None = None, tags: list[str] | None = None) -> str:
    """Find papers similar to a given one by its Zotero key or title.

    Args:
        key_or_title: Zotero item key (e.g., ABC12345) or title search query (e.g., "attention is all you need")
        limit: Maximum number of similar papers to return (default 10)
        collection: Filter by Zotero collection name
        tags: Filter by tags (OR logic)

    Returns:
        JSON array of similar papers, or candidate list if multiple title matches found.
    """
    config = _get_config()
    searcher, vector_store, bm25 = _init_searcher(config)

    key, candidates = _resolve_key_mcp(key_or_title, vector_store, bm25)
    if key is None and candidates:
        return json.dumps({"status": "multiple_matches", "query": key_or_title, "candidates": candidates}, indent=2)
    if key is None:
        return json.dumps({"status": "not_found", "query": key_or_title})

    tag_list = tags if tags else None
    results = searcher.similar_search(key, limit=limit, collection=collection, tags=tag_list)

    output = []
    for r in results:
        item = {"key": r.doc_id, "title": r.title, "score": round(r.score, 4), "source": r.source}
        if r.snippet:
            item["snippet"] = r.snippet
        output.append(item)

    return json.dumps(output, indent=2, ensure_ascii=False)


@mcp.tool()
def ask_papers(question: str, limit: int = 5, hybrid: bool = False,
               collection: str | None = None, tags: list[str] | None = None) -> str:
    """Ask a question and get an LLM-generated answer based on your indexed papers.

    Args:
        question: The question to answer
        limit: Number of papers to use as context (default 5)
        hybrid: Use hybrid search for context retrieval
        collection: Filter by Zotero collection name
        tags: Filter by tags (OR logic)

    Returns:
        JSON object with answer text and source papers used.
    """
    from rak.llm import LLMClient

    config = _get_config()
    searcher, _, _ = _init_searcher(config)

    tag_list = tags if tags else None
    if hybrid:
        results = searcher.hybrid_search(question, limit=limit, collection=collection, tags=tag_list)
    else:
        results = searcher.vector_search(question, limit=limit, collection=collection, tags=tag_list)

    if not results:
        return json.dumps({"answer": "No relevant papers found.", "sources": []})

    context = [{"key": r.doc_id, "title": r.title, "text": r.snippet, "score": r.score} for r in results]

    llm = LLMClient(base_url=config.llm_base_url, model=config.llm_model, api_key=config.llm_api_key)
    answer = llm.ask(question, context)

    return json.dumps({
        "answer": answer,
        "sources": [{"key": c["key"], "title": c["title"], "score": round(c["score"], 4)} for c in context],
    }, indent=2, ensure_ascii=False)


@mcp.tool()
def export_papers(query: str, limit: int = 10, format: str = "csv",
                  hybrid: bool = False, collection: str | None = None,
                  tags: list[str] | None = None) -> str:
    """Export search results as CSV or BibTeX.

    Args:
        query: Search query
        limit: Maximum number of results (default 10)
        format: Export format, either "csv" or "bibtex" (default "csv")
        hybrid: Use hybrid search
        collection: Filter by Zotero collection name
        tags: Filter by tags (OR logic)

    Returns:
        Formatted CSV or BibTeX string.
    """
    from rak.export import to_bibtex, to_csv

    config = _get_config()
    searcher, vector_store, _ = _init_searcher(config)

    tag_list = tags if tags else None
    if hybrid:
        results = searcher.hybrid_search(query, limit=limit, collection=collection, tags=tag_list)
    else:
        results = searcher.vector_search(query, limit=limit, collection=collection, tags=tag_list)

    if not results:
        return "No results found."

    # Batch fetch metadata
    meta_map: dict[str, dict] = {}
    if vector_store is not None:
        all_ids = [r.doc_id for r in results]
        doc_data = vector_store.get(ids=all_ids, include=["metadatas"])
        for doc_id, meta in zip(doc_data["ids"], doc_data["metadatas"]):
            meta_map[doc_id] = meta

    export_rows = []
    for r in results:
        meta = meta_map.get(r.doc_id, {})
        export_rows.append({
            "key": r.doc_id, "title": r.title, "score": r.score, "source": r.source,
            "date": meta.get("date", ""), "authors": meta.get("authors", ""),
            "item_type": meta.get("item_type", ""),
        })

    if format == "bibtex":
        return to_bibtex(export_rows)
    return to_csv(export_rows)


@mcp.tool()
def show_config() -> str:
    """Show current rak configuration values.

    Returns:
        JSON object with all configuration key-value pairs.
    """
    config = _get_config()
    return json.dumps({
        "model_name": config.model_name,
        "zot_command": config.zot_command,
        "llm_base_url": config.llm_base_url,
        "llm_model": config.llm_model,
        "embedding_provider": config.embedding_provider,
        "embedding_base_url": config.embedding_base_url if config.embedding_provider == "api" else None,
        "pdf_provider": config.pdf_provider,
        "chunk_size": config.chunk_size,
        "chunk_overlap": config.chunk_overlap,
        "data_dir": str(config.data_dir),
        "zotero_storage_dir": str(config.zotero_storage_dir) if config.zotero_storage_dir else None,
    }, indent=2)


@mcp.tool()
def index_status() -> str:
    """Show the current index status including item count, model, and last indexed time.

    Returns:
        JSON object with index metadata or a message if no index exists.
    """
    from rak.metadata import load_metadata

    config = _get_config()
    meta = load_metadata(config.data_dir)

    if meta is None:
        return json.dumps({"status": "No index found. Run 'rak index' first."})

    return json.dumps({
        "item_count": meta.item_count,
        "model_name": meta.model_name,
        "last_indexed": meta.last_indexed,
        "data_directory": str(config.data_dir),
    }, indent=2)


def main():
    mcp.run()


if __name__ == "__main__":
    main()



from pathlib import Path
from unittest.mock import patch, MagicMock

from rak.indexer import diff_items


def test_diff_items_passes_pdf_provider(tmp_path: Path):
    """diff_items passes pdf_provider to extract_file_text."""
    items = [{"key": "ABC", "title": "Test Paper"}]
    registry = {}
    storage_dir = tmp_path
    key_dir = storage_dir / "ABC"
    key_dir.mkdir()
    (key_dir / "paper.pdf").write_bytes(b"%PDF")

    with patch("rak.indexer.extract_file_text", return_value="extracted") as mock_extract:
        diff_items(items, registry, storage_dir=storage_dir, pdf_provider="mineru")

    mock_extract.assert_called()
    call_kwargs = mock_extract.call_args
    assert call_kwargs[1].get("provider") == "mineru" or \
           (len(call_kwargs[0]) > 1 and call_kwargs[0][1] == "mineru")



from __future__ import annotations

import hashlib
import json
from pathlib import Path

REGISTRY_FILENAME = "registry.json"


def compute_hash(text: str) -> str:
    return hashlib.md5(text.encode()).hexdigest()


def save_registry(data_dir: Path, registry: dict[str, str]) -> None:
    import os
    import tempfile
    target = data_dir / REGISTRY_FILENAME
    fd, tmp_path = tempfile.mkstemp(dir=data_dir, suffix=".tmp")
    try:
        os.write(fd, json.dumps(registry).encode())
        os.close(fd)
        os.replace(tmp_path, target)
    except BaseException:
        try:
            os.close(fd)
        except OSError:
            pass
        if os.path.exists(tmp_path):
            os.unlink(tmp_path)
        raise


def load_registry(data_dir: Path) -> dict[str, str]:
    path = data_dir / REGISTRY_FILENAME
    if not path.exists():
        return {}
    try:
        data = json.loads(path.read_text())
    except (json.JSONDecodeError, OSError):
        return {}
    if not isinstance(data, dict):
        return {}
    # Ensure all keys and values are strings
    return {str(k): str(v) for k, v in data.items()}



from pathlib import Path
from unittest.mock import patch

import pytest

from rak.config import RakConfig, detect_zotero_storage, save_config, load_config


def test_detect_zotero_storage_found(tmp_path: Path):
    storage_dir = tmp_path / "Zotero" / "storage"
    storage_dir.mkdir(parents=True)
    with patch("rak.config.Path.home", return_value=tmp_path):
        result = detect_zotero_storage()
    assert result == storage_dir


def test_detect_zotero_storage_missing(tmp_path: Path):
    with patch("rak.config.Path.home", return_value=tmp_path):
        result = detect_zotero_storage()
    assert result is None


def test_rak_config_has_zotero_storage():
    config = RakConfig()
    assert config.zotero_storage_dir is None or isinstance(config.zotero_storage_dir, Path)


def test_save_and_load_config(tmp_path: Path):
    save_config(tmp_path, "llm_model", "mistral")
    cfg = load_config(tmp_path)
    assert cfg["llm_model"] == "mistral"


def test_save_config_merges(tmp_path: Path):
    save_config(tmp_path, "llm_model", "mistral")
    save_config(tmp_path, "llm_base_url", "http://localhost:1234/v1")
    cfg = load_config(tmp_path)
    assert cfg["llm_model"] == "mistral"
    assert cfg["llm_base_url"] == "http://localhost:1234/v1"


def test_load_config_missing(tmp_path: Path):
    cfg = load_config(tmp_path)
    assert cfg == {}


def test_save_config_coerces_chunk_size_to_int(tmp_path: Path):
    save_config(tmp_path, "chunk_size", "256")
    cfg = load_config(tmp_path)
    assert cfg["chunk_size"] == 256
    assert isinstance(cfg["chunk_size"], int)


def test_save_config_rejects_invalid_chunk_size(tmp_path: Path):
    with pytest.raises(ValueError, match="expected int"):
        save_config(tmp_path, "chunk_size", "not_a_number")


def test_save_config_rejects_malicious_zot_command(tmp_path: Path):
    with pytest.raises(ValueError, match="Invalid zot_command"):
        save_config(tmp_path, "zot_command", "/usr/bin/evil")


def test_save_config_accepts_simple_zot_command(tmp_path: Path):
    save_config(tmp_path, "zot_command", "zot")
    cfg = load_config(tmp_path)
    assert cfg["zot_command"] == "zot"


def test_config_loads_chunk_size_as_int(tmp_path: Path):
    import json
    # Simulate legacy config with string chunk_size
    (tmp_path / "config.json").write_text(json.dumps({"chunk_size": "128"}))
    config = RakConfig(data_dir=tmp_path)
    assert config.chunk_size == 128
    assert isinstance(config.chunk_size, int)


def test_save_and_load_pdf_provider(tmp_path: Path):
    save_config(tmp_path, "pdf_provider", "mineru")
    cfg = load_config(tmp_path)
    assert cfg["pdf_provider"] == "mineru"


def test_config_loads_pdf_provider(tmp_path: Path):
    import json
    (tmp_path / "config.json").write_text(json.dumps({"pdf_provider": "mineru"}))
    config = RakConfig(data_dir=tmp_path)
    assert config.pdf_provider == "mineru"


def test_config_default_pdf_provider(tmp_path: Path):
    config = RakConfig(data_dir=tmp_path)
    assert config.pdf_provider == "pymupdf"


def test_save_config_rejects_invalid_pdf_provider(tmp_path: Path):
    with pytest.raises(ValueError, match="Invalid pdf_provider"):
        save_config(tmp_path, "pdf_provider", "invalid_provider")


def test_save_config_accepts_valid_pdf_providers(tmp_path: Path):
    for provider in ("pymupdf", "mineru", "docling"):
        save_config(tmp_path, "pdf_provider", provider)
        cfg = load_config(tmp_path)
        assert cfg["pdf_provider"] == provider



from unittest.mock import patch
from rak.errors import RakError, ZotNotFoundError, EmptyLibraryError, ModelDownloadError
from rak.indexer import fetch_zot_items
from rak.embedder import Embedder
import pytest


def test_zot_not_found_is_rak_error():
    err = ZotNotFoundError("zot")
    assert isinstance(err, RakError)
    assert "zot" in str(err)


def test_empty_library_is_rak_error():
    err = EmptyLibraryError()
    assert isinstance(err, RakError)


def test_model_download_error_is_rak_error():
    err = ModelDownloadError("all-MiniLM-L6-v2", "Connection refused")
    assert isinstance(err, RakError)
    assert "all-MiniLM-L6-v2" in str(err)
    assert "Connection refused" in str(err)


def test_fetch_zot_items_raises_zot_not_found():
    with patch("shutil.which", return_value=None):
        with pytest.raises(ZotNotFoundError):
            fetch_zot_items("zot")


def test_fetch_zot_items_raises_empty_library():
    with patch("shutil.which", return_value="/usr/bin/zot"), \
         patch("subprocess.run") as mock_run:
        mock_run.return_value.returncode = 0
        mock_run.return_value.stdout = "[]"
        with pytest.raises(EmptyLibraryError):
            fetch_zot_items("zot")


def test_embedder_raises_model_download_error():
    with patch("sentence_transformers.SentenceTransformer", side_effect=OSError("Connection refused")):
        with pytest.raises(ModelDownloadError, match="Connection refused"):
            Embedder("bad-model-name")



from unittest.mock import MagicMock, patch

import pytest

from rak.llm import LLMClient, LLMConnectionError


def test_ask_builds_prompt_and_returns_response():
    with patch("rak.llm.OpenAI") as mock_openai_cls:
        mock_client = mock_openai_cls.return_value
        mock_response = MagicMock()
        mock_response.choices = [MagicMock()]
        mock_response.choices[0].message.content = "The answer is 42."
        mock_client.chat.completions.create.return_value = mock_response

        client = LLMClient(base_url="http://localhost:11434/v1", model="llama3")
        context = [
            {"key": "A1", "title": "Paper One", "text": "Content of paper one."},
            {"key": "A2", "title": "Paper Two", "text": "Content of paper two."},
        ]
        answer = client.ask("What is the meaning?", context)

        assert answer == "The answer is 42."
        call_args = mock_client.chat.completions.create.call_args
        messages = call_args.kwargs["messages"]
        assert len(messages) == 2
        assert messages[0]["role"] == "system"
        assert "research" in messages[0]["content"].lower()
        assert messages[1]["role"] == "user"
        assert "What is the meaning?" in messages[1]["content"]
        assert "Paper One" in messages[1]["content"]
        assert "Paper Two" in messages[1]["content"]


def test_ask_connection_error():
    with patch("rak.llm.OpenAI") as mock_openai_cls:
        mock_client = mock_openai_cls.return_value
        from openai import APIConnectionError
        mock_client.chat.completions.create.side_effect = APIConnectionError(request=MagicMock())

        client = LLMClient(base_url="http://localhost:11434/v1", model="llama3")
        with pytest.raises(LLMConnectionError, match="not reachable"):
            client.ask("test", [])


def test_ask_stream_yields_tokens():
    with patch("rak.llm.OpenAI") as mock_openai_cls:
        mock_client = mock_openai_cls.return_value
        # Simulate streaming chunks
        chunks = []
        for text in ["The ", "answer ", "is ", "42."]:
            chunk = MagicMock()
            chunk.choices = [MagicMock()]
            chunk.choices[0].delta.content = text
            chunks.append(chunk)
        mock_client.chat.completions.create.return_value = iter(chunks)

        client = LLMClient(base_url="http://localhost:11434/v1", model="llama3")
        context = [{"key": "A1", "title": "Paper", "text": "Content."}]
        tokens = list(client.ask_stream("question", context))

        assert tokens == ["The ", "answer ", "is ", "42."]
        call_args = mock_client.chat.completions.create.call_args
        assert call_args.kwargs["stream"] is True


def test_ask_stream_connection_error():
    with patch("rak.llm.OpenAI") as mock_openai_cls:
        mock_client = mock_openai_cls.return_value
        from openai import APIConnectionError
        mock_client.chat.completions.create.side_effect = APIConnectionError(request=MagicMock())

        client = LLMClient(base_url="http://localhost:11434/v1", model="llama3")
        with pytest.raises(LLMConnectionError, match="not reachable"):
            list(client.ask_stream("test", []))


def test_ask_server_error():
    from rak.llm import LLMServerError
    with patch("rak.llm.OpenAI") as mock_openai_cls:
        mock_client = mock_openai_cls.return_value
        from openai import APIStatusError
        mock_response = MagicMock()
        mock_response.status_code = 500
        mock_response.json.return_value = {"error": "model not found"}
        mock_client.chat.completions.create.side_effect = APIStatusError(
            message="model not found", response=mock_response, body=None,
        )

        client = LLMClient(base_url="http://localhost:11434/v1", model="llama3")
        with pytest.raises(LLMServerError, match="LLM server error"):
            client.ask("test", [])


def test_ask_stream_server_error():
    from rak.llm import LLMServerError
    with patch("rak.llm.OpenAI") as mock_openai_cls:
        mock_client = mock_openai_cls.return_value
        from openai import APIStatusError
        mock_response = MagicMock()
        mock_response.status_code = 404
        mock_response.json.return_value = {}
        mock_client.chat.completions.create.side_effect = APIStatusError(
            message="not found", response=mock_response, body=None,
        )

        client = LLMClient(base_url="http://localhost:11434/v1", model="llama3")
        with pytest.raises(LLMServerError):
            list(client.ask_stream("test", []))


def test_ask_stream_empty_delta():
    """Chunks with no content should be skipped."""
    with patch("rak.llm.OpenAI") as mock_openai_cls:
        mock_client = mock_openai_cls.return_value
        chunk1 = MagicMock()
        chunk1.choices = [MagicMock()]
        chunk1.choices[0].delta.content = None  # no content
        chunk2 = MagicMock()
        chunk2.choices = [MagicMock()]
        chunk2.choices[0].delta.content = "Hello"
        chunk3 = MagicMock()
        chunk3.choices = []  # no choices
        mock_client.chat.completions.create.return_value = iter([chunk1, chunk2, chunk3])

        client = LLMClient(base_url="http://localhost:11434/v1", model="llama3")
        tokens = list(client.ask_stream("q", [{"key": "A", "title": "T", "text": "C"}]))
        assert tokens == ["Hello"]


def test_build_messages_format():
    """Verify message structure."""
    with patch("rak.llm.OpenAI"):
        client = LLMClient(base_url="http://localhost:11434/v1", model="llama3")
    context = [{"key": "A1", "title": "Paper", "text": "Content"}]
    messages = client._build_messages("What is X?", context)
    assert len(messages) == 2
    assert messages[0]["role"] == "system"
    assert messages[1]["role"] == "user"
    assert "[1] A1 - Paper" in messages[1]["content"]
    assert "What is X?" in messages[1]["content"]


def test_build_messages_empty_context():
    with patch("rak.llm.OpenAI"):
        client = LLMClient(base_url="http://localhost:11434/v1", model="llama3")
    messages = client._build_messages("question", [])
    assert "No papers found" in messages[1]["content"]



from pathlib import Path
from unittest.mock import patch, MagicMock

from rak.pdf import chunk_text, extract_pdf_text, extract_file_text, find_attachments, _split_paragraphs

FIXTURES = Path(__file__).parent / "fixtures"


def test_extract_pdf_text():
    text = extract_pdf_text(FIXTURES / "sample.pdf")
    assert "test PDF document" in text
    assert len(text) > 10


def test_extract_pdf_text_missing_file():
    text = extract_pdf_text(Path("/nonexistent/file.pdf"))
    assert text == ""


def test_find_attachments_pdf_and_md(tmp_path: Path):
    key_dir = tmp_path / "ABC12345"
    key_dir.mkdir()
    pdf_file = key_dir / "Smith 2024 - Paper.pdf"
    pdf_file.touch()
    md_file = key_dir / "notes.md"
    md_file.write_text("some notes")
    result = find_attachments(tmp_path, "ABC12345")
    assert len(result) == 2
    assert pdf_file in result
    assert md_file in result


def test_find_attachments_not_found(tmp_path: Path):
    result = find_attachments(tmp_path, "NONEXIST")
    assert result == []


def test_find_attachments_ignores_other_files(tmp_path: Path):
    key_dir = tmp_path / "ABC12345"
    key_dir.mkdir()
    (key_dir / "snapshot.html").touch()
    (key_dir / "image.png").touch()
    result = find_attachments(tmp_path, "ABC12345")
    assert result == []


def test_extract_file_text_md(tmp_path: Path):
    md = tmp_path / "note.md"
    md.write_text("# Title\nSome content here.")
    text = extract_file_text(md)
    assert "# Title" in text
    assert "Some content here." in text


def test_extract_file_text_unsupported(tmp_path: Path):
    txt = tmp_path / "readme.txt"
    txt.write_text("hello")
    assert extract_file_text(txt) == ""


def test_chunk_text_empty():
    assert chunk_text("") == []


def test_chunk_text_short():
    text = "hello world this is a short text"
    chunks = chunk_text(text, chunk_size=512)
    assert len(chunks) == 1
    assert chunks[0] == text


def test_chunk_text_splits():
    words = [f"word{i}" for i in range(100)]
    text = " ".join(words)
    chunks = chunk_text(text, chunk_size=30, overlap=10)
    assert len(chunks) > 1
    # Each chunk (except possibly last) should have ~30 words
    first_words = chunks[0].split()
    assert len(first_words) == 30
    # Overlap: second chunk should start 20 words in
    second_words = chunks[1].split()
    assert second_words[0] == "word20"


def test_chunk_text_overlap_content():
    words = [f"w{i}" for i in range(50)]
    text = " ".join(words)
    chunks = chunk_text(text, chunk_size=20, overlap=5)
    # Last words of chunk 0 should appear at start of chunk 1
    c0_words = chunks[0].split()
    c1_words = chunks[1].split()
    assert c0_words[-5:] == c1_words[:5]


def test_split_paragraphs_double_newline():
    text = "First paragraph.\n\nSecond paragraph.\n\nThird paragraph."
    paras = _split_paragraphs(text)
    assert len(paras) == 3
    assert paras[0] == "First paragraph."
    assert paras[1] == "Second paragraph."


def test_split_paragraphs_markdown_heading():
    text = "Intro text.\n\n## Methods\n\nMethod details.\n\n## Results\n\nResult details."
    paras = _split_paragraphs(text)
    assert any("Methods" in p for p in paras)
    assert any("Results" in p for p in paras)


def test_chunk_text_respects_paragraphs():
    """Paragraphs that fit within chunk_size should be merged, not split mid-sentence."""
    para1 = " ".join(f"a{i}" for i in range(10))  # 10 words
    para2 = " ".join(f"b{i}" for i in range(10))  # 10 words
    para3 = " ".join(f"c{i}" for i in range(10))  # 10 words
    para4 = " ".join(f"d{i}" for i in range(10))  # 10 words
    text = f"{para1}\n\n{para2}\n\n{para3}\n\n{para4}"
    chunks = chunk_text(text, chunk_size=25, overlap=5)
    # Should merge paras into 2 chunks (10+10=20 fits, 10+10=20 fits) not split mid-para
    assert len(chunks) == 2
    assert "a0" in chunks[0]
    assert "b0" in chunks[0]
    assert "c0" in chunks[1]
    assert "d0" in chunks[1]


def test_chunk_text_oversized_paragraph():
    """A single oversized paragraph should be split with word-level overlap."""
    big_para = " ".join(f"x{i}" for i in range(50))
    small_para = "short paragraph here"
    text = f"{big_para}\n\n{small_para}"
    chunks = chunk_text(text, chunk_size=20, overlap=5)
    assert len(chunks) >= 3  # big para splits into ~3 chunks + small para
    assert "short paragraph here" in chunks[-1]


def test_chunk_text_no_paragraph_structure():
    """Plain text without paragraph breaks falls back to word-level chunking."""
    words = [f"w{i}" for i in range(100)]
    text = " ".join(words)
    chunks = chunk_text(text, chunk_size=30, overlap=10)
    assert len(chunks) > 1
    first_words = chunks[0].split()
    assert len(first_words) == 30


def test_extract_pdf_text_mineru_provider(tmp_path: Path):
    """MinerU provider calls subprocess and reads markdown output."""
    pdf_file = tmp_path / "paper.pdf"
    pdf_file.write_bytes(b"%PDF-1.4 fake")

    def fake_run(cmd, **kwargs):
        # Simulate mineru writing a markdown file
        # The output dir is the last argument
        output_dir = Path(cmd[-1])
        output_dir.mkdir(parents=True, exist_ok=True)
        auto_dir = output_dir / pdf_file.stem / "auto"
        auto_dir.mkdir(parents=True, exist_ok=True)
        md_file = auto_dir / (pdf_file.stem + ".md")
        md_file.write_text("# Parsed Title\n\nTable | Data\n---|---\n1 | 2")
        result = MagicMock()
        result.returncode = 0
        return result

    with patch("rak.pdf.subprocess.run", side_effect=fake_run):
        text = extract_pdf_text(pdf_file, provider="mineru")

    assert "Parsed Title" in text
    assert "Table" in text


def test_extract_pdf_text_mineru_fallback_on_failure(tmp_path: Path):
    """When MinerU fails, falls back to PyMuPDF."""
    pdf_file = FIXTURES / "sample.pdf"

    def fake_run(cmd, **kwargs):
        result = MagicMock()
        result.returncode = 1
        result.stderr = "mineru not found"
        return result

    with patch("rak.pdf.subprocess.run", side_effect=fake_run):
        text = extract_pdf_text(pdf_file, provider="mineru")

    # Should fall back to PyMuPDF and extract text
    assert "test PDF document" in text


def test_extract_pdf_text_mineru_fallback_on_exception(tmp_path: Path):
    """When MinerU subprocess raises, falls back to PyMuPDF."""
    pdf_file = FIXTURES / "sample.pdf"

    with patch("rak.pdf.subprocess.run", side_effect=FileNotFoundError("mineru not found")):
        text = extract_pdf_text(pdf_file, provider="mineru")

    assert "test PDF document" in text


def test_extract_file_text_passes_provider(tmp_path: Path):
    """extract_file_text passes provider to extract_pdf_text."""
    pdf_file = tmp_path / "test.pdf"
    pdf_file.write_bytes(b"%PDF-1.4 fake")

    with patch("rak.pdf.extract_pdf_text", return_value="mineru text") as mock:
        text = extract_file_text(pdf_file, provider="mineru")

    mock.assert_called_once_with(pdf_file, provider="mineru")
    assert text == "mineru text"


def test_extract_pdf_text_docling_provider(tmp_path: Path):
    """Docling provider calls subprocess and reads markdown output."""
    pdf_file = tmp_path / "paper.pdf"
    pdf_file.write_bytes(b"%PDF-1.4 fake")

    def fake_run(cmd, **kwargs):
        # Docling outputs markdown to the --output dir
        output_idx = cmd.index("--output") + 1
        output_dir = Path(cmd[output_idx])
        output_dir.mkdir(parents=True, exist_ok=True)
        md_file = output_dir / (pdf_file.stem + ".md")
        md_file.write_text("# Docling Title\n\n| Col1 | Col2 |\n|---|---|\n| a | b |")
        result = MagicMock()
        result.returncode = 0
        return result

    with patch("rak.pdf.subprocess.run", side_effect=fake_run):
        text = extract_pdf_text(pdf_file, provider="docling")

    assert "Docling Title" in text
    assert "Col1" in text


def test_extract_pdf_text_docling_fallback_on_failure(tmp_path: Path):
    """When Docling fails, falls back to PyMuPDF."""
    pdf_file = FIXTURES / "sample.pdf"

    def fake_run(cmd, **kwargs):
        result = MagicMock()
        result.returncode = 1
        result.stderr = "docling error"
        return result

    with patch("rak.pdf.subprocess.run", side_effect=fake_run):
        text = extract_pdf_text(pdf_file, provider="docling")

    assert "test PDF document" in text


def test_extract_pdf_text_docling_fallback_on_exception(tmp_path: Path):
    """When Docling subprocess raises, falls back to PyMuPDF."""
    pdf_file = FIXTURES / "sample.pdf"

    with patch("rak.pdf.subprocess.run", side_effect=FileNotFoundError("docling not found")):
        text = extract_pdf_text(pdf_file, provider="docling")

    assert "test PDF document" in text



import json
from pathlib import Path

from rak.metadata import IndexMetadata, save_metadata, load_metadata


def test_save_and_load_roundtrip(tmp_path: Path):
    save_metadata(tmp_path, model_name="all-MiniLM-L6-v2", item_count=342)
    meta = load_metadata(tmp_path)
    assert meta is not None
    assert meta.model_name == "all-MiniLM-L6-v2"
    assert meta.item_count == 342
    assert meta.last_indexed  # non-empty ISO timestamp


def test_load_missing_returns_none(tmp_path: Path):
    assert load_metadata(tmp_path) is None


def test_load_corrupt_json_returns_none(tmp_path: Path):
    (tmp_path / "meta.json").write_text("not valid json {{{")
    assert load_metadata(tmp_path) is None


def test_load_missing_keys_returns_none(tmp_path: Path):
    (tmp_path / "meta.json").write_text(json.dumps({"model_name": "test"}))
    assert load_metadata(tmp_path) is None


def test_save_metadata_atomic_write(tmp_path: Path):
    """save_metadata uses atomic write — no leftover tmp files on success."""
    save_metadata(tmp_path, model_name="test-model", item_count=10)
    meta_path = tmp_path / "meta.json"
    assert meta_path.exists()
    data = json.loads(meta_path.read_text())
    assert data["model_name"] == "test-model"
    assert data["item_count"] == 10
    assert "last_indexed" in data
    # No leftover .tmp files
    tmp_files = list(tmp_path.glob("*.tmp"))
    assert tmp_files == []


def test_save_metadata_overwrites_existing(tmp_path: Path):
    save_metadata(tmp_path, model_name="model-v1", item_count=100)
    save_metadata(tmp_path, model_name="model-v2", item_count=200)
    meta = load_metadata(tmp_path)
    assert meta.model_name == "model-v2"
    assert meta.item_count == 200



from __future__ import annotations

import json
from io import StringIO

from rich.console import Console
from rich.table import Table

from rak.searcher import SearchResult


def format_results(results: list[SearchResult], output_json: bool = False) -> str:
    if output_json:
        items = []
        for r in results:
            item = {"key": r.doc_id, "title": r.title, "score": round(r.score, 4), "source": r.source}
            if r.snippet:
                item["snippet"] = r.snippet
            items.append(item)
        return json.dumps(items, indent=2, ensure_ascii=False)
    buf = StringIO()
    console = Console(file=buf, force_terminal=False, width=120)
    table = Table(show_header=True, header_style="bold")
    table.add_column("Key", style="cyan", width=10)
    table.add_column("Title", width=60)
    table.add_column("Score", width=8, justify="right")
    table.add_column("Source", width=10)
    for r in results:
        table.add_row(r.doc_id, r.title, f"{r.score:.3f}", r.source)
    console.print(table)
    return buf.getvalue()


def format_index_stats(count: int, output_json: bool = False) -> str:
    if output_json:
        return json.dumps({"indexed": count})
    return f"Indexed {count} papers."


def format_incremental_stats(stats: dict, output_json: bool = False) -> str:
    if output_json:
        return json.dumps({k: v for k, v in stats.items() if k not in ("registry", "text_cache")})
    parts = []
    if stats["added"]:
        parts.append(f"{stats['added']} new")
    if stats["updated"]:
        parts.append(f"{stats['updated']} updated")
    if stats["removed"]:
        parts.append(f"{stats['removed']} removed")
    parts.append(f"{stats['unchanged']} unchanged")
    return ", ".join(parts) + "."


def format_ask_result(
    answer: str,
    sources: list[dict],
    output_json: bool = False,
) -> str:
    if output_json:
        return json.dumps({
            "answer": answer,
            "sources": [
                {"key": s["key"], "title": s["title"], "score": round(s["score"], 4)}
                for s in sources
            ],
        }, indent=2, ensure_ascii=False)
    lines = [answer, "", "Sources:"]
    for i, s in enumerate(sources, 1):
        lines.append(f"  {i}. {s['key']} - {s['title']} (score: {s['score']:.3f})")
    return "\n".join(lines)



from __future__ import annotations

import json
import logging
import os
import shutil
import subprocess

from concurrent.futures import ThreadPoolExecutor, as_completed
from pathlib import Path

logger = logging.getLogger(__name__)

_MAX_EXTRACT_WORKERS = min(8, os.cpu_count() or 4)

from rak.bm25 import BM25Index
from rak.embedder import Embedder
from rak.errors import EmptyLibraryError, ZotNotFoundError
from rak.pdf import chunk_text, extract_file_text, find_attachments
from rak.registry import compute_hash
from rak.store import VectorStore


def parse_zot_items(raw_json: str) -> list[dict]:
    return json.loads(raw_json)


def build_document_text(item: dict, pdf_text: str = "") -> str:
    parts = [item.get("title", "")]
    creators = item.get("creators", [])
    if creators:
        author_names = []
        for c in creators:
            name_parts = [c.get("first_name", ""), c.get("last_name", "")]
            author_names.append(" ".join(p for p in name_parts if p))
        parts.append("Authors: " + ", ".join(author_names))
    abstract = item.get("abstract")
    if abstract:
        parts.append(abstract)
    tags = item.get("tags", [])
    if tags:
        parts.append("Tags: " + ", ".join(tags))
    if pdf_text:
        parts.append(pdf_text)
    return "\n".join(parts)


def _extract_item_text(
    item: dict, storage_dir: Path | None, pdf_provider: str,
) -> tuple[str, int, int]:
    """Extract attachment text for a single item (thread-safe).

    Returns (attachment_text, extraction_attempts, extraction_failures).
    """
    key = item.get("key", "")
    if not key or not storage_dir:
        return "", 0, 0
    attachments = find_attachments(storage_dir, key)
    if not attachments:
        return "", 0, 0
    attempts = len(attachments)
    failures = 0
    texts = []
    for p in attachments:
        t = extract_file_text(p, provider=pdf_provider)
        if t:
            texts.append(t)
        else:
            failures += 1
    return "\n\n".join(texts), attempts, failures


def diff_items(
    items: list[dict], registry: dict[str, str],
    storage_dir: Path | None = None,
    pdf_provider: str = "pymupdf",
) -> tuple[list[dict], list[dict], list[str], dict[str, str]]:
    """Compare items against registry to find changes.

    Returns (to_add, to_update, to_remove, text_cache) where text_cache
    maps item keys to their built document text, avoiding redundant PDF extraction.
    """
    to_add = []
    to_update = []
    text_cache: dict[str, str] = {}
    fetched_keys = set()
    extraction_attempts = 0
    extraction_failures = 0

    # Phase 1: parallel PDF/attachment extraction
    attachment_texts: dict[str, str] = {}
    valid_items = [(item, item.get("key", "")) for item in items]
    valid_items = [(item, key) for item, key in valid_items if key]

    if storage_dir:
        with ThreadPoolExecutor(max_workers=_MAX_EXTRACT_WORKERS) as executor:
            future_to_key = {
                executor.submit(_extract_item_text, item, storage_dir, pdf_provider): key
                for item, key in valid_items
            }
            for future in as_completed(future_to_key):
                key = future_to_key[future]
                att_text, att_attempts, att_failures = future.result()
                extraction_attempts += att_attempts
                extraction_failures += att_failures
                if att_text:
                    attachment_texts[key] = att_text

    # Phase 2: build documents and diff against registry
    for item, key in valid_items:
        fetched_keys.add(key)
        text = build_document_text(item, pdf_text=attachment_texts.get(key, ""))
        if not text.strip():
            continue
        text_cache[key] = text
        content_hash = compute_hash(text)
        if key not in registry:
            to_add.append(item)
        elif registry[key] != content_hash:
            to_update.append(item)

    if extraction_attempts > 0 and extraction_failures > 0:
        rate = extraction_failures / extraction_attempts
        if rate > 0.5:
            logger.warning(
                "%d of %d file extractions failed (%.0f%%). Check PDF provider '%s'.",
                extraction_failures, extraction_attempts, rate * 100, pdf_provider,
            )
        elif extraction_failures > 0:
            logger.info("%d of %d file extractions failed.", extraction_failures, extraction_attempts)
    to_remove = [k for k in registry if k not in fetched_keys]
    return to_add, to_update, to_remove, text_cache


def fetch_zot_items(zot_command: str = "zot", limit: int = 5000) -> list[dict]:
    if not shutil.which(zot_command):
        raise ZotNotFoundError(zot_command)
    result = subprocess.run(
        [zot_command, "--json", "--limit", str(limit), "list"],
        capture_output=True, text=True, encoding="utf-8",
    )
    if result.returncode != 0:
        raise RuntimeError(f"zot command failed: {result.stderr}")
    items = parse_zot_items(result.stdout)
    if not items:
        raise EmptyLibraryError()
    return items


def index_items(
    items: list[dict],
    embedder: Embedder,
    vector_store: VectorStore,
    bm25_index: BM25Index,
    on_progress: callable | None = None,
    registry: dict[str, str] | None = None,
    storage_dir: Path | None = None,
    chunk_size: int = 512,
    chunk_overlap: int = 64,
    pdf_provider: str = "pymupdf",
) -> dict:
    """Index items and return a result dict.

    Returns dict with keys: added, updated, removed, unchanged, registry, text_cache.
    For full indexing (registry=None), added=count and updated/removed/unchanged=0.
    """
    if registry is not None:
        return _index_incremental(items, embedder, vector_store, bm25_index, on_progress, registry, storage_dir, chunk_size, chunk_overlap, pdf_provider)
    return _index_full(items, embedder, vector_store, bm25_index, on_progress, storage_dir, chunk_size, chunk_overlap, pdf_provider)


def _delete_chunks(vector_store: VectorStore, key: str) -> None:
    """Delete a document and any associated chunks from the vector store."""
    # Try deleting the base document
    vector_store.delete([key])
    # Delete any chunks (query by parent_key metadata)
    try:
        results = vector_store.get_ids_by_metadata(where={"parent_key": key})
        if results["ids"]:
            vector_store.delete(results["ids"])
    except Exception as exc:
        logger.debug("Failed to delete chunks for %s: %s", key, exc)


def _build_metadata(item: dict) -> dict:
    collections = item.get("collections", [])
    tags = item.get("tags", [])
    metadata = {
        "title": item.get("title", ""),
        "date": item.get("date", ""),
        "item_type": item.get("item_type", ""),
    }
    creators = item.get("creators", [])
    if creators:
        author_parts = [
            " ".join(p for p in [c.get("first_name", ""), c.get("last_name", "")] if p)
            for c in creators
        ]
        metadata["authors"] = ", ".join(author_parts)
    if collections:
        metadata["collections"] = collections
    if tags:
        metadata["tags"] = tags
    return metadata


def _embed_and_store_batch(
    batch_ids: list[str],
    batch_texts: list[str],
    batch_metadatas: list[dict],
    embedder: Embedder,
    vector_store: VectorStore,
) -> None:
    """Embed a batch of texts and store them in the vector store."""
    if not batch_texts:
        return
    embeddings = embedder.embed_batch(batch_texts)
    vector_store.add(ids=batch_ids, embeddings=embeddings, documents=batch_texts, metadatas=batch_metadatas)


def _index_full(
    items: list[dict],
    embedder: Embedder,
    vector_store: VectorStore,
    bm25_index: BM25Index,
    on_progress: callable | None = None,
    storage_dir: Path | None = None,
    chunk_size: int = 512,
    chunk_overlap: int = 64,
    pdf_provider: str = "pymupdf",
) -> dict:
    """Index all items from scratch. Returns result dict."""
    count = 0
    text_cache: dict[str, str] = {}
    batch_ids: list[str] = []
    batch_texts: list[str] = []
    batch_metadatas: list[dict] = []
    embed_batch_size = embedder.batch_size
    extraction_attempts = 0
    extraction_failures = 0

    # Phase 1: parallel PDF/attachment extraction
    valid_items = [(item, item.get("key", "")) for item in items]
    valid_items = [(item, key) for item, key in valid_items if key]

    attachment_texts: dict[str, str] = {}
    if storage_dir:
        with ThreadPoolExecutor(max_workers=_MAX_EXTRACT_WORKERS) as executor:
            future_to_key = {
                executor.submit(_extract_item_text, item, storage_dir, pdf_provider): key
                for item, key in valid_items
            }
            for future in as_completed(future_to_key):
                key = future_to_key[future]
                att_text, att_attempts, att_failures = future.result()
                extraction_attempts += att_attempts
                extraction_failures += att_failures
                if att_text:
                    attachment_texts[key] = att_text

    # Phase 2: sequential embedding and indexing
    for i, (item, key) in enumerate(valid_items):
        text = build_document_text(item, attachment_texts.get(key, ""))
        if not text.strip():
            continue
        text_cache[key] = text
        metadata = _build_metadata(item)
        # Store full document for BM25
        bm25_index.add(key, text)
        # Accumulate chunks for batch embedding
        chunks = chunk_text(text, chunk_size=chunk_size, overlap=chunk_overlap)
        if len(chunks) <= 1:
            batch_ids.append(key)
            batch_texts.append(text)
            batch_metadatas.append(metadata)
        else:
            for ci, chunk in enumerate(chunks):
                chunk_id = f"{key}_chunk_{ci}"
                chunk_meta = {**metadata, "parent_key": key, "chunk_index": ci}
                batch_ids.append(chunk_id)
                batch_texts.append(chunk)
                batch_metadatas.append(chunk_meta)
        # Flush batch when full
        if len(batch_texts) >= embed_batch_size:
            _embed_and_store_batch(batch_ids, batch_texts, batch_metadatas, embedder, vector_store)
            batch_ids, batch_texts, batch_metadatas = [], [], []
        count += 1
        if on_progress:
            on_progress(i + 1, len(valid_items))

    # Flush remaining
    _embed_and_store_batch(batch_ids, batch_texts, batch_metadatas, embedder, vector_store)
    if extraction_attempts > 0 and extraction_failures > 0:
        rate = extraction_failures / extraction_attempts
        if rate > 0.5:
            logger.warning(
                "%d of %d file extractions failed (%.0f%%). Check PDF provider '%s'.",
                extraction_failures, extraction_attempts, rate * 100, pdf_provider,
            )
        elif extraction_failures > 0:
            logger.info("%d of %d file extractions failed.", extraction_failures, extraction_attempts)
    return {
        "added": count,
        "updated": 0,
        "removed": 0,
        "unchanged": 0,
        "registry": {k: compute_hash(v) for k, v in text_cache.items()},
        "text_cache": text_cache,
    }


def _index_incremental(
    items: list[dict],
    embedder: Embedder,
    vector_store: VectorStore,
    bm25_index: BM25Index,
    on_progress: callable | None = None,
    registry: dict[str, str] | None = None,
    storage_dir: Path | None = None,
    chunk_size: int = 512,
    chunk_overlap: int = 64,
    pdf_provider: str = "pymupdf",
) -> dict[str, int]:
    to_add, to_update, to_remove, text_cache = diff_items(items, registry, storage_dir=storage_dir, pdf_provider=pdf_provider)
    new_registry = dict(registry)
    added = 0
    updated = 0

    work_items = [(item, "add") for item in to_add] + [(item, "update") for item in to_update]
    batch_ids: list[str] = []
    batch_texts: list[str] = []
    batch_metadatas: list[dict] = []
    embed_batch_size = embedder.batch_size

    for i, (item, action) in enumerate(work_items):
        key = item["key"]
        text = text_cache.get(key, "")
        if not text.strip():
            continue
        metadata = _build_metadata(item)
        if action == "update":
            bm25_index.delete(key)
            _delete_chunks(vector_store, key)
        bm25_index.add(key, text)
        chunks = chunk_text(text, chunk_size=chunk_size, overlap=chunk_overlap)
        if len(chunks) <= 1:
            batch_ids.append(key)
            batch_texts.append(text)
            batch_metadatas.append(metadata)
        else:
            for ci, chunk in enumerate(chunks):
                chunk_id = f"{key}_chunk_{ci}"
                chunk_meta = {**metadata, "parent_key": key, "chunk_index": ci}
                batch_ids.append(chunk_id)
                batch_texts.append(chunk)
                batch_metadatas.append(chunk_meta)
        if len(batch_texts) >= embed_batch_size:
            _embed_and_store_batch(batch_ids, batch_texts, batch_metadatas, embedder, vector_store)
            batch_ids, batch_texts, batch_metadatas = [], [], []
        new_registry[key] = compute_hash(text)
        if action == "add":
            added += 1
        else:
            updated += 1
        if on_progress:
            on_progress(i + 1, len(work_items))

    _embed_and_store_batch(batch_ids, batch_texts, batch_metadatas, embedder, vector_store)

    for key in to_remove:
        _delete_chunks(vector_store, key)
        bm25_index.delete(key)
        del new_registry[key]

    return {
        "added": added,
        "updated": updated,
        "removed": len(to_remove),
        "unchanged": len(items) - len(to_add) - len(to_update),
        "registry": new_registry,
        "text_cache": text_cache,
    }



from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("zotero-rag-cli")
except PackageNotFoundError:
    __version__ = "0.7.4"  # fallback for editable installs without metadata



import pytest

from rak.bm25 import BM25Index


@pytest.fixture
def bm25(tmp_path):
    return BM25Index(tmp_path / "fts.sqlite")


def test_add_and_search(bm25):
    bm25.add("A1", "single cell RNA sequencing analysis")
    bm25.add("A2", "deep learning for image classification")
    results = bm25.search("single cell RNA", limit=5)
    assert len(results) >= 1
    assert results[0]["id"] == "A1"


def test_search_ranking(bm25):
    bm25.add("A1", "RNA RNA RNA sequencing single cell")
    bm25.add("A2", "RNA processing")
    results = bm25.search("RNA", limit=2)
    assert results[0]["id"] == "A1"


def test_search_no_results(bm25):
    bm25.add("A1", "deep learning")
    results = bm25.search("quantum physics", limit=5)
    assert len(results) == 0


def test_count(bm25):
    assert bm25.count() == 0
    bm25.add("A1", "test document")
    assert bm25.count() == 1


def test_clear(bm25):
    bm25.add("A1", "test")
    bm25.clear()
    assert bm25.count() == 0


def test_search_returns_score(bm25):
    bm25.add("A1", "transformer attention mechanism")
    results = bm25.search("transformer", limit=1)
    assert "score" in results[0]
    assert results[0]["score"] > 0


def test_delete(bm25):
    bm25.add("A1", "paper about RNA")
    bm25.add("A2", "paper about DNA")
    assert bm25.count() == 2
    bm25.delete("A1")
    assert bm25.count() == 1


def test_search_query_with_quotes(bm25):
    bm25.add("A1", "he said hello world")
    results = bm25.search('he said "hello"', limit=5)
    assert isinstance(results, list)


def test_search_with_snippet(bm25):
    bm25.add("A1", "single cell RNA sequencing analysis method")
    bm25.add("A2", "deep learning for image classification")
    results = bm25.search_with_snippet("single cell RNA", limit=5)
    assert len(results) >= 1
    assert results[0]["id"] == "A1"
    assert "snippet" in results[0]
    assert results[0]["snippet"]  # non-empty


def test_search_with_snippet_no_results(bm25):
    bm25.add("A1", "deep learning")
    results = bm25.search_with_snippet("quantum physics", limit=5)
    assert results == []


def test_search_with_snippet_has_title_key(bm25):
    bm25.add("A1", "transformer attention mechanism")
    results = bm25.search_with_snippet("transformer", limit=1)
    assert "title" in results[0]



import pytest

from rak.store import VectorStore


@pytest.fixture
def store(tmp_path):
    return VectorStore(persist_dir=tmp_path / "chroma", dimension=384)


def test_add_and_search(store):
    store.add(
        ids=["A1", "A2"],
        embeddings=[[1.0] + [0.0] * 383, [0.0] + [1.0] + [0.0] * 382],
        documents=["paper about RNA", "paper about DNA"],
        metadatas=[{"title": "RNA paper"}, {"title": "DNA paper"}],
    )
    results = store.search(query_embedding=[1.0] + [0.0] * 383, limit=1)
    assert len(results) == 1
    assert results[0]["id"] == "A1"


def test_search_returns_score(store):
    store.add(
        ids=["X1"],
        embeddings=[[1.0] + [0.0] * 383],
        documents=["test"],
        metadatas=[{"title": "test"}],
    )
    results = store.search(query_embedding=[1.0] + [0.0] * 383, limit=1)
    assert "score" in results[0]
    assert results[0]["score"] > 0


def test_count(store):
    assert store.count() == 0
    store.add(
        ids=["Z1"],
        embeddings=[[0.5] * 384],
        documents=["doc"],
        metadatas=[{"title": "doc"}],
    )
    assert store.count() == 1


def test_clear(store):
    store.add(
        ids=["Z1"],
        embeddings=[[0.5] * 384],
        documents=["doc"],
        metadatas=[{"title": "doc"}],
    )
    store.clear()
    assert store.count() == 0


def test_has(store):
    assert not store.has("Z1")
    store.add(
        ids=["Z1"],
        embeddings=[[0.5] * 384],
        documents=["doc"],
        metadatas=[{"title": "doc"}],
    )
    assert store.has("Z1")


def test_delete(store):
    store.add(
        ids=["A1", "A2"],
        embeddings=[[1.0] + [0.0] * 383, [0.0] + [1.0] + [0.0] * 382],
        documents=["paper one", "paper two"],
        metadatas=[{"title": "one"}, {"title": "two"}],
    )
    assert store.count() == 2
    store.delete(["A1"])
    assert store.count() == 1
    assert not store.has("A1")
    assert store.has("A2")


def test_search_with_where_filter(store):
    store.add(
        ids=["A1", "A2"],
        embeddings=[[1.0] + [0.0] * 383, [0.9] + [0.1] + [0.0] * 382],
        documents=["paper about RNA", "paper about DNA"],
        metadatas=[
            {"title": "RNA paper", "collections": ["Biology", "Genomics"], "tags": ["RNA", "seq"]},
            {"title": "DNA paper", "collections": ["Chemistry"], "tags": ["DNA"]},
        ],
    )
    results = store.search(
        query_embedding=[1.0] + [0.0] * 383,
        limit=10,
        where={"collections": {"$contains": "Biology"}},
    )
    assert len(results) == 1
    assert results[0]["id"] == "A1"


def test_get_embedding(store):
    emb = [1.0] + [0.0] * 383
    store.add(ids=["E1"], embeddings=[emb], documents=["doc"], metadatas=[{"title": "t"}])
    result = store.get_embedding("E1")
    assert result is not None
    assert len(result) == 384
    assert result[0] == pytest.approx(1.0, abs=0.01)


def test_get_embedding_missing(store):
    result = store.get_embedding("NONEXIST")
    assert result is None



"""Tests for rak.formatter — format_results, format_index_stats,
format_incremental_stats, format_ask_result."""

import json

from rak.formatter import (
    format_ask_result,
    format_incremental_stats,
    format_index_stats,
    format_results,
)
from rak.searcher import SearchResult


# --- format_results ---


def test_format_results_json():
    results = [
        SearchResult(doc_id="A1", score=0.9512, title="Paper One", source="vector", snippet="chunk text"),
        SearchResult(doc_id="A2", score=0.7, title="Paper Two", source="bm25"),
    ]
    out = format_results(results, output_json=True)
    data = json.loads(out)
    assert len(data) == 2
    assert data[0]["key"] == "A1"
    assert data[0]["title"] == "Paper One"
    assert data[0]["score"] == 0.9512
    assert data[0]["source"] == "vector"
    assert data[0]["snippet"] == "chunk text"
    # No snippet for second result
    assert "snippet" not in data[1]


def test_format_results_json_empty():
    out = format_results([], output_json=True)
    assert json.loads(out) == []


def test_format_results_table():
    results = [
        SearchResult(doc_id="X1", score=0.95, title="RNA Sequencing", source="vector"),
    ]
    out = format_results(results, output_json=False)
    assert "X1" in out
    assert "RNA Sequencing" in out
    assert "0.950" in out
    assert "vector" in out


def test_format_results_table_empty():
    out = format_results([], output_json=False)
    # Rich table with no rows should still have header or be mostly empty
    assert isinstance(out, str)


# --- format_index_stats ---


def test_format_index_stats_json():
    out = format_index_stats(42, output_json=True)
    data = json.loads(out)
    assert data == {"indexed": 42}


def test_format_index_stats_text():
    out = format_index_stats(42, output_json=False)
    assert out == "Indexed 42 papers."


# --- format_incremental_stats ---


def test_format_incremental_stats_json():
    stats = {"added": 5, "updated": 2, "removed": 1, "unchanged": 10, "registry": {}, "text_cache": {}}
    out = format_incremental_stats(stats, output_json=True)
    data = json.loads(out)
    assert data["added"] == 5
    assert data["updated"] == 2
    assert data["removed"] == 1
    assert data["unchanged"] == 10
    # registry and text_cache should be excluded
    assert "registry" not in data
    assert "text_cache" not in data


def test_format_incremental_stats_text_all_changes():
    stats = {"added": 3, "updated": 2, "removed": 1, "unchanged": 10}
    out = format_incremental_stats(stats, output_json=False)
    assert "3 new" in out
    assert "2 updated" in out
    assert "1 removed" in out
    assert "10 unchanged" in out
    assert out.endswith(".")


def test_format_incremental_stats_text_only_added():
    stats = {"added": 5, "updated": 0, "removed": 0, "unchanged": 0}
    out = format_incremental_stats(stats, output_json=False)
    assert "5 new" in out
    assert "updated" not in out
    assert "removed" not in out
    assert "0 unchanged" in out


def test_format_incremental_stats_text_no_changes():
    stats = {"added": 0, "updated": 0, "removed": 0, "unchanged": 50}
    out = format_incremental_stats(stats, output_json=False)
    assert "50 unchanged" in out
    assert "new" not in out
    assert "updated" not in out
    assert "removed" not in out


# --- format_ask_result ---


def test_format_ask_result_json():
    sources = [
        {"key": "A1", "title": "Paper One", "score": 0.91234},
        {"key": "A2", "title": "Paper Two", "score": 0.71111},
    ]
    out = format_ask_result("The answer is 42.", sources, output_json=True)
    data = json.loads(out)
    assert data["answer"] == "The answer is 42."
    assert len(data["sources"]) == 2
    assert data["sources"][0]["key"] == "A1"
    assert data["sources"][0]["score"] == 0.9123  # rounded to 4 decimals
    assert data["sources"][1]["score"] == 0.7111


def test_format_ask_result_text():
    sources = [
        {"key": "A1", "title": "Paper One", "score": 0.912},
        {"key": "B2", "title": "Paper Two", "score": 0.700},
    ]
    out = format_ask_result("Deep learning is great.", sources, output_json=False)
    assert "Deep learning is great." in out
    assert "Sources:" in out
    assert "1. A1 - Paper One (score: 0.912)" in out
    assert "2. B2 - Paper Two (score: 0.700)" in out


def test_format_ask_result_text_no_sources():
    out = format_ask_result("No context available.", [], output_json=False)
    assert "No context available." in out
    assert "Sources:" in out


def test_format_ask_result_json_empty_sources():
    out = format_ask_result("answer", [], output_json=True)
    data = json.loads(out)
    assert data["sources"] == []



"""Custom exception hierarchy for rak."""
from __future__ import annotations


class RakError(Exception):
    """Base exception for rak."""


class ZotNotFoundError(RakError):
    """Raised when zot command is not found on PATH."""

    def __init__(self, command: str = "zot") -> None:
        super().__init__(f"'{command}' not found on PATH. Install: pip install zotero-cli-cc")


class EmptyLibraryError(RakError):
    """Raised when Zotero library contains no items."""

    def __init__(self) -> None:
        super().__init__("No items found in Zotero library.")


class DimensionMismatchError(RakError):
    """Raised when embedding dimension doesn't match existing index."""

    def __init__(self, expected: int, got: int) -> None:
        super().__init__(
            f"Embedding dimension mismatch: index has {expected}D vectors but current model produces {got}D. "
            f"Run 'rak reindex' to rebuild with the new model."
        )


class ModelDownloadError(RakError):
    """Raised when model download or loading fails."""

    def __init__(self, model_name: str, reason: str = "") -> None:
        msg = f"Failed to load model '{model_name}'"
        if reason:
            msg += f": {reason}"
        super().__init__(msg)



import pytest

from rak.embedder import Embedder


@pytest.fixture
def embedder():
    return Embedder(model_name="all-MiniLM-L6-v2")


@pytest.mark.network
def test_embed_single(embedder):
    vec = embedder.embed("hello world")
    assert isinstance(vec, list)
    assert len(vec) == 384
    assert all(isinstance(x, float) for x in vec)


@pytest.mark.network
def test_embed_batch(embedder):
    texts = ["paper about RNA", "deep learning methods", "cell biology"]
    vecs = embedder.embed_batch(texts)
    assert len(vecs) == 3
    assert all(len(v) == 384 for v in vecs)


@pytest.mark.network
def test_embed_empty_string(embedder):
    vec = embedder.embed("")
    assert len(vec) == 384


@pytest.mark.network
def test_model_name(embedder):
    assert embedder.model_name == "all-MiniLM-L6-v2"


@pytest.mark.network
def test_dimension(embedder):
    assert embedder.dimension == 384


# --- API provider tests (no network needed) ---

from unittest.mock import patch, MagicMock


def test_api_provider_embed():
    mock_resp = MagicMock()
    mock_resp.data = [MagicMock()]
    mock_resp.data[0].embedding = [0.1, 0.2, 0.3]

    with patch("openai.OpenAI") as mock_openai:
        mock_client = mock_openai.return_value
        mock_client.embeddings.create.return_value = mock_resp
        emb = Embedder(model_name="text-embedding-3-small", provider="api",
                       base_url="http://localhost:11434/v1", api_key="test-key")
        vec = emb.embed("hello world")

    assert vec == [0.1, 0.2, 0.3]
    mock_client.embeddings.create.assert_called_once_with(
        input=["hello world"], model="text-embedding-3-small"
    )


def test_api_provider_embed_batch():
    mock_resp = MagicMock()
    d0 = MagicMock()
    d0.embedding = [0.1, 0.2]
    d0.index = 0
    d1 = MagicMock()
    d1.embedding = [0.3, 0.4]
    d1.index = 1
    mock_resp.data = [d1, d0]  # intentionally out of order

    with patch("openai.OpenAI") as mock_openai:
        mock_client = mock_openai.return_value
        mock_client.embeddings.create.return_value = mock_resp
        emb = Embedder(model_name="text-embedding-3-small", provider="api")
        vecs = emb.embed_batch(["text1", "text2"])

    assert vecs == [[0.1, 0.2], [0.3, 0.4]]  # sorted by index


def test_api_provider_dimension():
    mock_resp = MagicMock()
    mock_resp.data = [MagicMock()]
    mock_resp.data[0].embedding = [0.0] * 1536

    with patch("openai.OpenAI") as mock_openai:
        mock_client = mock_openai.return_value
        mock_client.embeddings.create.return_value = mock_resp
        emb = Embedder(model_name="text-embedding-3-small", provider="api")
        dim = emb.dimension

    assert dim == 1536


def test_api_provider_model_name():
    with patch("openai.OpenAI"):
        emb = Embedder(model_name="custom-model", provider="api")
    assert emb.model_name == "custom-model"



from __future__ import annotations

from pathlib import Path

import chromadb

from rak.errors import DimensionMismatchError


class VectorStore:
    COLLECTION_NAME = "rak_papers"

    def __init__(self, persist_dir: Path, dimension: int = 384) -> None:
        self._client = chromadb.PersistentClient(path=str(persist_dir))
        self._collection = self._client.get_or_create_collection(
            name=self.COLLECTION_NAME,
            metadata={"hnsw:space": "cosine"},
        )
        self._dimension = dimension
        self._validate_dimension()

    def _validate_dimension(self) -> None:
        """Check that stored embeddings match the expected dimension."""
        if self._collection.count() == 0:
            return
        sample = self._collection.get(limit=1, include=["embeddings"])
        if sample["embeddings"] is not None and len(sample["embeddings"]) > 0:
            stored_dim = len(sample["embeddings"][0])
            if stored_dim != self._dimension:
                raise DimensionMismatchError(expected=stored_dim, got=self._dimension)

    def add(self, ids: list[str], embeddings: list[list[float]], documents: list[str], metadatas: list[dict]) -> None:
        self._collection.upsert(ids=ids, embeddings=embeddings, documents=documents, metadatas=metadatas)

    def delete(self, ids: list[str]) -> None:
        self._collection.delete(ids=ids)

    def search(self, query_embedding: list[float], limit: int = 10, where: dict | None = None) -> list[dict]:
        total = self._collection.count()
        if total == 0:
            return []
        kwargs = {
            "query_embeddings": [query_embedding],
            "n_results": min(limit, total),
            "include": ["documents", "metadatas", "distances"],
        }
        if where is not None:
            kwargs["where"] = where
        results = self._collection.query(**kwargs)
        items = []
        for i in range(len(results["ids"][0])):
            items.append({
                "id": results["ids"][0][i],
                "document": results["documents"][0][i],
                "metadata": results["metadatas"][0][i],
                "score": max(0.0, 1.0 - results["distances"][0][i]),
            })
        return items

    def count(self) -> int:
        return self._collection.count()

    def get(self, ids: list[str], include: list[str] | None = None) -> dict:
        """Retrieve documents/metadatas by IDs."""
        return self._collection.get(ids=ids, include=include or ["documents"])

    def get_ids_by_metadata(self, where: dict) -> dict:
        """Query document IDs by metadata filter."""
        return self._collection.get(where=where, include=[])

    def get_embedding(self, doc_id: str) -> list[float] | None:
        """Retrieve the embedding vector for a document ID."""
        result = self._collection.get(ids=[doc_id], include=["embeddings"])
        if not result["ids"]:
            return None
        return result["embeddings"][0]

    def has(self, doc_id: str) -> bool:
        result = self._collection.get(ids=[doc_id], include=[])
        return len(result["ids"]) > 0

    def clear(self) -> None:
        self._client.delete_collection(self.COLLECTION_NAME)
        self._collection = self._client.get_or_create_collection(
            name=self.COLLECTION_NAME, metadata={"hnsw:space": "cosine"},
        )



from unittest.mock import patch, MagicMock

from click.testing import CliRunner

from rak.cli import main


def test_version():
    runner = CliRunner()
    result = runner.invoke(main, ["--version"])
    from rak import __version__
    assert __version__ in result.output


def test_help():
    runner = CliRunner()
    result = runner.invoke(main, ["--help"])
    assert "Semantic search" in result.output
    assert result.exit_code == 0


def test_search_help():
    runner = CliRunner()
    result = runner.invoke(main, ["search", "--help"])
    assert "--hybrid" in result.output
    assert "--limit" in result.output


def test_index_help():
    runner = CliRunner()
    result = runner.invoke(main, ["index", "--help"])
    assert result.exit_code == 0


def test_index_zot_not_found():
    runner = CliRunner()
    with patch("rak.indexer.shutil.which", return_value=None):
        result = runner.invoke(main, ["index"])
    assert result.exit_code == 1
    assert "not found" in result.output


def test_index_empty_library():
    runner = CliRunner()
    with patch("rak.indexer.shutil.which", return_value="/usr/bin/zot"), \
         patch("rak.indexer.subprocess.run") as mock_run, \
         patch("sentence_transformers.SentenceTransformer"):
        mock_run.return_value.returncode = 0
        mock_run.return_value.stdout = "[]"
        result = runner.invoke(main, ["index"])
    assert result.exit_code == 0
    assert "No items found" in result.output


def test_search_model_download_error():
    runner = CliRunner()
    with patch("sentence_transformers.SentenceTransformer", side_effect=OSError("Network error")):
        result = runner.invoke(main, ["search", "test query"])
    assert result.exit_code == 1
    assert "Failed to load model" in result.output


import json as json_mod
from pathlib import Path
from rak.metadata import save_metadata


def test_index_writes_metadata(tmp_path: Path):
    from rak.config import RakConfig
    runner = CliRunner()
    zot_item = json_mod.dumps([{
        "key": "ABC123",
        "title": "Test Paper",
        "creators": [],
        "abstract": "An abstract.",
        "date": "2024",
        "tags": [],
        "item_type": "journalArticle",
    }])
    fake_config = RakConfig(data_dir=tmp_path)
    with patch("rak.indexer.shutil.which", return_value="/usr/bin/zot"), \
         patch("rak.indexer.subprocess.run") as mock_run, \
         patch("sentence_transformers.SentenceTransformer") as mock_st, \
         patch("rak.cli.RakConfig", return_value=fake_config):
        mock_run.return_value.returncode = 0
        mock_run.return_value.stdout = zot_item
        mock_model = mock_st.return_value
        import numpy as np
        def _fake_encode(text_or_texts, **kwargs):
            if isinstance(text_or_texts, list):
                return np.array([[0.1] * 384 for _ in text_or_texts])
            return np.array([0.1] * 384)
        mock_model.encode.side_effect = _fake_encode
        mock_model.get_sentence_embedding_dimension.return_value = 384
        result = runner.invoke(main, ["index"])

    meta_path = tmp_path / "meta.json"
    assert meta_path.exists(), f"meta.json not found; output was: {result.output}"
    meta = json_mod.loads(meta_path.read_text())
    assert meta["item_count"] >= 1
    assert meta["model_name"] == "all-MiniLM-L6-v2"


def test_status_no_index(tmp_path: Path):
    from rak.config import RakConfig
    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()
    with patch("rak.cli.RakConfig", return_value=fake_config):
        result = runner.invoke(main, ["status"])
    assert result.exit_code == 0
    assert "No index found" in result.output


def test_status_with_index(tmp_path: Path):
    from rak.config import RakConfig
    save_metadata(tmp_path, model_name="all-MiniLM-L6-v2", item_count=342)
    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()
    with patch("rak.cli.RakConfig", return_value=fake_config):
        result = runner.invoke(main, ["status"])
    assert result.exit_code == 0
    assert "342" in result.output
    assert "all-MiniLM-L6-v2" in result.output


def test_status_json_output(tmp_path: Path):
    from rak.config import RakConfig
    save_metadata(tmp_path, model_name="all-MiniLM-L6-v2", item_count=100)
    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()
    with patch("rak.cli.RakConfig", return_value=fake_config):
        result = runner.invoke(main, ["--json", "status"])
    data = json_mod.loads(result.output)
    assert data["item_count"] == 100
    assert data["model_name"] == "all-MiniLM-L6-v2"


def test_clear_with_yes(tmp_path: Path):
    from rak.config import RakConfig
    # Create fake index files
    (tmp_path / "chroma").mkdir()
    (tmp_path / "fts.sqlite").touch()
    save_metadata(tmp_path, model_name="test", item_count=10)

    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()
    with patch("rak.cli.RakConfig", return_value=fake_config):
        result = runner.invoke(main, ["clear", "--yes"])
    assert result.exit_code == 0
    assert "Cleared" in result.output
    assert not (tmp_path / "chroma").exists()
    assert not (tmp_path / "fts.sqlite").exists()
    assert not (tmp_path / "meta.json").exists()


def test_clear_nothing_to_clear(tmp_path: Path):
    from rak.config import RakConfig
    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()
    with patch("rak.cli.RakConfig", return_value=fake_config):
        result = runner.invoke(main, ["clear", "--yes"])
    assert result.exit_code == 0
    assert "Nothing to clear" in result.output


def test_clear_prompts_without_yes(tmp_path: Path):
    from rak.config import RakConfig
    (tmp_path / "chroma").mkdir()
    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()
    with patch("rak.cli.RakConfig", return_value=fake_config):
        result = runner.invoke(main, ["clear"], input="n\n")
    assert (tmp_path / "chroma").exists()  # not deleted because user said no


def test_index_help_shows_full_flag():
    runner = CliRunner()
    result = runner.invoke(main, ["index", "--help"])
    assert "--full" in result.output


def test_search_help_shows_filter_flags():
    runner = CliRunner()
    result = runner.invoke(main, ["search", "--help"])
    assert "--collection" in result.output
    assert "--tag" in result.output


def test_ask_help():
    runner = CliRunner()
    result = runner.invoke(main, ["ask", "--help"])
    assert result.exit_code == 0
    assert "--context" in result.output
    assert "--hybrid" in result.output
    assert "--bm25" in result.output
    assert "--llm-model" in result.output
    assert "--llm-url" in result.output
    assert "--collection" in result.output
    assert "--tag" in result.output


def test_config_help():
    runner = CliRunner()
    result = runner.invoke(main, ["config", "--help"])
    assert result.exit_code == 0
    assert "Show or set" in result.output


def test_export_help():
    runner = CliRunner()
    result = runner.invoke(main, ["export", "--help"])
    assert result.exit_code == 0
    assert "--format" in result.output
    assert "--output" in result.output
    assert "--hybrid" in result.output
    assert "--bm25" in result.output
    assert "--collection" in result.output
    assert "--tag" in result.output


def test_chat_help():
    runner = CliRunner()
    result = runner.invoke(main, ["chat", "--help"])
    assert result.exit_code == 0
    assert "Interactive multi-turn" in result.output
    assert "--context" in result.output
    assert "--hybrid" in result.output
    assert "--bm25" in result.output
    assert "--llm-model" in result.output


def test_completion_help():
    runner = CliRunner()
    result = runner.invoke(main, ["completion", "--help"])
    assert result.exit_code == 0
    assert "shell completion" in result.output.lower()


# --- Config integration tests ---


def test_config_show_all(tmp_path: Path):
    from rak.config import RakConfig
    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()
    with patch("rak.cli.RakConfig", return_value=fake_config):
        result = runner.invoke(main, ["config"])
    assert result.exit_code == 0
    assert "model_name = " in result.output
    assert "zot_command = " in result.output
    assert "llm_base_url = " in result.output
    assert "llm_model = " in result.output
    assert "llm_api_key = " in result.output


def test_config_get_single_key(tmp_path: Path):
    from rak.config import RakConfig
    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()
    with patch("rak.cli.RakConfig", return_value=fake_config):
        result = runner.invoke(main, ["config", "llm_model"])
    assert result.exit_code == 0
    assert "llm_model = llama3" in result.output


def test_config_get_unknown_key(tmp_path: Path):
    from rak.config import RakConfig
    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()
    with patch("rak.cli.RakConfig", return_value=fake_config):
        result = runner.invoke(main, ["config", "nonexistent_key"])
    assert result.exit_code == 1


def test_config_set_value(tmp_path: Path):
    from rak.config import RakConfig, load_config
    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()
    with patch("rak.cli.RakConfig", return_value=fake_config):
        result = runner.invoke(main, ["config", "llm_model", "mistral"])
    assert result.exit_code == 0
    assert "llm_model = mistral" in result.output
    cfg = load_config(tmp_path)
    assert cfg["llm_model"] == "mistral"


def test_config_set_unknown_key(tmp_path: Path):
    from rak.config import RakConfig
    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()
    with patch("rak.cli.RakConfig", return_value=fake_config):
        result = runner.invoke(main, ["config", "bad_key", "value"])
    assert result.exit_code == 1
    assert "Unknown config key" in result.output


def test_config_masks_api_key(tmp_path: Path):
    from rak.config import RakConfig, save_config
    save_config(tmp_path, "llm_api_key", "sk-abcdefghijklmnop")
    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()
    with patch("rak.cli.RakConfig", return_value=fake_config):
        result = runner.invoke(main, ["config"])
    assert "..." in result.output
    assert "sk-abcdefghijklmnop" not in result.output


# --- Export integration tests ---


def test_export_csv_output(tmp_path: Path):
    from rak.config import RakConfig
    from rak.searcher import SearchResult

    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()

    mock_results = [
        SearchResult(doc_id="A1", score=0.9, title="Paper One", source="vector"),
        SearchResult(doc_id="A2", score=0.7, title="Paper Two", source="vector"),
    ]

    with patch("rak.cli.RakConfig", return_value=fake_config), \
         patch("sentence_transformers.SentenceTransformer") as mock_st, \
         patch("rak.searcher.Searcher.vector_search", return_value=mock_results), \
         patch("rak.store.VectorStore.__init__", return_value=None), \
         patch("rak.store.VectorStore.get", return_value={"ids": ["A1", "A2"], "metadatas": [{"date": "2024", "authors": "Doe"}, {"date": "2024", "authors": "Doe"}]}), \
         patch("rak.bm25.BM25Index.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.close"):
        mock_st.return_value.get_sentence_embedding_dimension.return_value = 384
        result = runner.invoke(main, ["export", "test query"])

    assert result.exit_code == 0
    assert "key,title,score,source" in result.output
    assert "A1" in result.output
    assert "Paper One" in result.output


def test_export_bibtex_output(tmp_path: Path):
    from rak.config import RakConfig
    from rak.searcher import SearchResult

    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()

    mock_results = [
        SearchResult(doc_id="B1", score=0.8, title="Deep Learning", source="vector"),
    ]

    with patch("rak.cli.RakConfig", return_value=fake_config), \
         patch("sentence_transformers.SentenceTransformer") as mock_st, \
         patch("rak.searcher.Searcher.vector_search", return_value=mock_results), \
         patch("rak.store.VectorStore.__init__", return_value=None), \
         patch("rak.store.VectorStore.get", return_value={"ids": ["B1"], "metadatas": [{"date": "2024", "authors": "Smith"}]}), \
         patch("rak.bm25.BM25Index.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.close"):
        mock_st.return_value.get_sentence_embedding_dimension.return_value = 384
        result = runner.invoke(main, ["export", "test query", "--format", "bibtex"])

    assert result.exit_code == 0
    assert "@article{B1," in result.output
    assert "Deep Learning" in result.output


def test_export_to_file(tmp_path: Path):
    from rak.config import RakConfig
    from rak.searcher import SearchResult

    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()
    output_file = str(tmp_path / "results.csv")

    mock_results = [
        SearchResult(doc_id="C1", score=0.5, title="Paper", source="vector"),
    ]

    with patch("rak.cli.RakConfig", return_value=fake_config), \
         patch("sentence_transformers.SentenceTransformer") as mock_st, \
         patch("rak.searcher.Searcher.vector_search", return_value=mock_results), \
         patch("rak.store.VectorStore.__init__", return_value=None), \
         patch("rak.store.VectorStore.get", return_value={"ids": ["C1"], "metadatas": [{}]}), \
         patch("rak.bm25.BM25Index.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.close"):
        mock_st.return_value.get_sentence_embedding_dimension.return_value = 384
        result = runner.invoke(main, ["export", "test", "--output", output_file])

    assert result.exit_code == 0
    assert "Exported" in result.output
    content = Path(output_file).read_text()
    assert "C1" in content


def test_export_no_results(tmp_path: Path):
    from rak.config import RakConfig

    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()

    with patch("rak.cli.RakConfig", return_value=fake_config), \
         patch("sentence_transformers.SentenceTransformer") as mock_st, \
         patch("rak.searcher.Searcher.vector_search", return_value=[]), \
         patch("rak.store.VectorStore.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.close"):
        mock_st.return_value.get_sentence_embedding_dimension.return_value = 384
        result = runner.invoke(main, ["export", "nonexistent"])

    assert result.exit_code == 0
    assert "No results found" in result.output


# --- Search integration tests ---


def test_search_vector_results(tmp_path: Path):
    from rak.config import RakConfig
    from rak.searcher import SearchResult

    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()

    mock_results = [
        SearchResult(doc_id="X1", score=0.95, title="RNA Sequencing", source="vector", snippet="chunk text"),
    ]

    with patch("rak.cli.RakConfig", return_value=fake_config), \
         patch("sentence_transformers.SentenceTransformer") as mock_st, \
         patch("rak.searcher.Searcher.vector_search", return_value=mock_results), \
         patch("rak.store.VectorStore.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.close"):
        mock_st.return_value.get_sentence_embedding_dimension.return_value = 384
        result = runner.invoke(main, ["search", "RNA"])

    assert result.exit_code == 0
    assert "RNA Sequencing" in result.output


def test_search_json_output(tmp_path: Path):
    from rak.config import RakConfig
    from rak.searcher import SearchResult

    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()

    mock_results = [
        SearchResult(doc_id="X1", score=0.95, title="Paper", source="vector", snippet="text"),
    ]

    with patch("rak.cli.RakConfig", return_value=fake_config), \
         patch("sentence_transformers.SentenceTransformer") as mock_st, \
         patch("rak.searcher.Searcher.vector_search", return_value=mock_results), \
         patch("rak.store.VectorStore.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.close"):
        mock_st.return_value.get_sentence_embedding_dimension.return_value = 384
        result = runner.invoke(main, ["--json", "search", "test"])

    assert result.exit_code == 0
    data = json_mod.loads(result.output)
    assert len(data) == 1
    assert data[0]["key"] == "X1"
    assert data[0]["snippet"] == "text"


def test_search_no_results_json(tmp_path: Path):
    from rak.config import RakConfig

    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()

    with patch("rak.cli.RakConfig", return_value=fake_config), \
         patch("sentence_transformers.SentenceTransformer") as mock_st, \
         patch("rak.searcher.Searcher.vector_search", return_value=[]), \
         patch("rak.store.VectorStore.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.close"):
        mock_st.return_value.get_sentence_embedding_dimension.return_value = 384
        result = runner.invoke(main, ["--json", "search", "nothing"])

    assert result.exit_code == 0
    data = json_mod.loads(result.output)
    assert data == []


def test_search_bm25_only(tmp_path: Path):
    from rak.config import RakConfig
    from rak.searcher import SearchResult

    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()

    mock_results = [
        SearchResult(doc_id="K1", score=5.0, title="", source="bm25", snippet="keyword match"),
    ]

    with patch("rak.cli.RakConfig", return_value=fake_config), \
         patch("rak.searcher.Searcher.bm25_search", return_value=mock_results), \
         patch("rak.bm25.BM25Index.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.close"):
        result = runner.invoke(main, ["search", "test", "--bm25"])

    assert result.exit_code == 0
    assert "K1" in result.output


def test_search_bm25_help_flag():
    runner = CliRunner()
    result = runner.invoke(main, ["search", "--help"])
    assert "--bm25" in result.output


def test_search_hybrid_mode(tmp_path: Path):
    from rak.config import RakConfig
    from rak.searcher import SearchResult

    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()

    mock_results = [
        SearchResult(doc_id="H1", score=0.03, title="Hybrid Paper", source="fused"),
    ]

    with patch("rak.cli.RakConfig", return_value=fake_config), \
         patch("sentence_transformers.SentenceTransformer") as mock_st, \
         patch("rak.searcher.Searcher.hybrid_search", return_value=mock_results), \
         patch("rak.store.VectorStore.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.close"):
        mock_st.return_value.get_sentence_embedding_dimension.return_value = 384
        result = runner.invoke(main, ["search", "test", "--hybrid"])

    assert result.exit_code == 0
    assert "Hybrid Paper" in result.output


# --- similar command tests ---


def test_similar_help():
    runner = CliRunner()
    result = runner.invoke(main, ["similar", "--help"])
    assert result.exit_code == 0
    assert "KEY_OR_TITLE" in result.output
    assert "--limit" in result.output


def test_similar_by_key(tmp_path: Path):
    from rak.config import RakConfig
    from rak.searcher import SearchResult

    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()

    mock_results = [
        SearchResult(doc_id="B1", score=0.85, title="Similar Paper", source="similar"),
    ]

    with patch("rak.cli.RakConfig", return_value=fake_config), \
         patch("sentence_transformers.SentenceTransformer") as mock_st, \
         patch("rak.store.VectorStore.__init__", return_value=None), \
         patch("rak.store.VectorStore.has", return_value=True), \
         patch("rak.searcher.Searcher.similar_search", return_value=mock_results), \
         patch("rak.bm25.BM25Index.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.close"):
        mock_st.return_value.get_sentence_embedding_dimension.return_value = 384
        result = runner.invoke(main, ["similar", "ABC12345"])

    assert result.exit_code == 0
    assert "Similar Paper" in result.output


def test_similar_no_results(tmp_path: Path):
    from rak.config import RakConfig

    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()

    with patch("rak.cli.RakConfig", return_value=fake_config), \
         patch("sentence_transformers.SentenceTransformer") as mock_st, \
         patch("rak.store.VectorStore.__init__", return_value=None), \
         patch("rak.store.VectorStore.has", return_value=True), \
         patch("rak.searcher.Searcher.similar_search", return_value=[]), \
         patch("rak.bm25.BM25Index.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.close"):
        mock_st.return_value.get_sentence_embedding_dimension.return_value = 384
        result = runner.invoke(main, ["similar", "ABC12345"])

    assert result.exit_code == 0
    assert "No similar papers" in result.output


# --- reindex command tests ---


def test_reindex_help():
    runner = CliRunner()
    result = runner.invoke(main, ["reindex", "--help"])
    assert result.exit_code == 0
    assert "Clear indexes and rebuild" in result.output


def test_reindex_zot_not_found(tmp_path: Path):
    from rak.config import RakConfig

    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()

    with patch("rak.cli.RakConfig", return_value=fake_config), \
         patch("sentence_transformers.SentenceTransformer") as mock_st, \
         patch("rak.indexer.shutil.which", return_value=None):
        mock_st.return_value.get_sentence_embedding_dimension.return_value = 384
        result = runner.invoke(main, ["reindex"])

    assert result.exit_code == 1
    assert "not found" in result.output


def test_reindex_empty_library(tmp_path: Path):
    from rak.config import RakConfig

    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()

    with patch("rak.cli.RakConfig", return_value=fake_config), \
         patch("sentence_transformers.SentenceTransformer") as mock_st, \
         patch("rak.indexer.shutil.which", return_value="/usr/bin/zot"), \
         patch("rak.indexer.subprocess.run") as mock_run:
        mock_st.return_value.get_sentence_embedding_dimension.return_value = 384
        mock_run.return_value.returncode = 0
        mock_run.return_value.stdout = "[]"
        result = runner.invoke(main, ["reindex"])

    assert result.exit_code == 0
    assert "No items found" in result.output


# --- ask command tests ---


def test_ask_json_output(tmp_path: Path):
    from rak.config import RakConfig
    from rak.searcher import SearchResult

    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()

    mock_results = [
        SearchResult(doc_id="D1", score=0.9, title="Context Paper", source="vector", snippet="some context"),
    ]

    with patch("rak.cli.RakConfig", return_value=fake_config), \
         patch("sentence_transformers.SentenceTransformer") as mock_st, \
         patch("rak.searcher.Searcher.vector_search", return_value=mock_results), \
         patch("rak.store.VectorStore.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.close"), \
         patch("rak.llm.OpenAI") as mock_openai:
        mock_st.return_value.get_sentence_embedding_dimension.return_value = 384
        mock_client = mock_openai.return_value
        mock_resp = MagicMock()
        mock_resp.choices = [MagicMock()]
        mock_resp.choices[0].message.content = "The answer."
        mock_client.chat.completions.create.return_value = mock_resp
        result = runner.invoke(main, ["--json", "ask", "What is it?"])

    assert result.exit_code == 0
    data = json_mod.loads(result.output)
    assert data["answer"] == "The answer."
    assert len(data["sources"]) == 1


def test_ask_streaming_output(tmp_path: Path):
    from rak.config import RakConfig
    from rak.searcher import SearchResult

    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()

    mock_results = [
        SearchResult(doc_id="D1", score=0.9, title="Context Paper", source="vector", snippet="ctx"),
    ]

    # Build streaming chunks
    chunks = []
    for text in ["Hello", " world"]:
        chunk = MagicMock()
        chunk.choices = [MagicMock()]
        chunk.choices[0].delta.content = text
        chunks.append(chunk)

    with patch("rak.cli.RakConfig", return_value=fake_config), \
         patch("sentence_transformers.SentenceTransformer") as mock_st, \
         patch("rak.searcher.Searcher.vector_search", return_value=mock_results), \
         patch("rak.store.VectorStore.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.close"), \
         patch("rak.llm.OpenAI") as mock_openai:
        mock_st.return_value.get_sentence_embedding_dimension.return_value = 384
        mock_client = mock_openai.return_value
        mock_client.chat.completions.create.return_value = iter(chunks)
        result = runner.invoke(main, ["ask", "What is it?"])

    assert result.exit_code == 0
    assert "Sources:" in result.output
    assert "D1" in result.output


def test_ask_no_results(tmp_path: Path):
    from rak.config import RakConfig

    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()

    with patch("rak.cli.RakConfig", return_value=fake_config), \
         patch("sentence_transformers.SentenceTransformer") as mock_st, \
         patch("rak.searcher.Searcher.vector_search", return_value=[]), \
         patch("rak.store.VectorStore.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.close"):
        mock_st.return_value.get_sentence_embedding_dimension.return_value = 384
        result = runner.invoke(main, ["ask", "anything"])

    assert result.exit_code == 0
    assert "No relevant papers" in result.output


def test_ask_llm_connection_error(tmp_path: Path):
    from rak.config import RakConfig
    from rak.searcher import SearchResult

    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()

    mock_results = [
        SearchResult(doc_id="D1", score=0.9, title="Paper", source="vector", snippet="ctx"),
    ]

    with patch("rak.cli.RakConfig", return_value=fake_config), \
         patch("sentence_transformers.SentenceTransformer") as mock_st, \
         patch("rak.searcher.Searcher.vector_search", return_value=mock_results), \
         patch("rak.store.VectorStore.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.close"), \
         patch("rak.llm.OpenAI") as mock_openai:
        mock_st.return_value.get_sentence_embedding_dimension.return_value = 384
        from openai import APIConnectionError
        mock_client = mock_openai.return_value
        mock_client.chat.completions.create.side_effect = APIConnectionError(request=MagicMock())
        result = runner.invoke(main, ["--json", "ask", "test"])

    assert result.exit_code == 1
    assert "not reachable" in result.output


# --- _resolve_key tests ---

from rak.cli import _resolve_key


def test_resolve_key_direct_match():
    mock_vs = MagicMock()
    mock_vs.has.return_value = True
    mock_bm25 = MagicMock()

    result = _resolve_key("ABC12345", mock_vs, mock_bm25)
    assert result == "ABC12345"
    mock_bm25.search.assert_not_called()


def test_resolve_key_not_found_in_store():
    """Key pattern but not in store -> falls back to title search."""
    mock_vs = MagicMock()
    mock_vs.has.return_value = False  # not found for either key or chunk_0
    mock_bm25 = MagicMock()
    mock_bm25.search.return_value = []

    runner = CliRunner()
    with runner.isolated_filesystem():
        result = _resolve_key("ZZZZZZZZ", mock_vs, mock_bm25)
    assert result is None


def test_resolve_key_title_single_match():
    """Title search returning one result -> auto-select."""
    mock_vs = MagicMock()
    mock_vs.has.return_value = False
    mock_vs.get.return_value = {
        "ids": ["KEY1"],
        "metadatas": [{"title": "Attention Is All You Need"}],
    }
    mock_bm25 = MagicMock()
    mock_bm25.search.return_value = [{"id": "KEY1", "score": 5.0}]

    result = _resolve_key("attention is all you need", mock_vs, mock_bm25)
    assert result == "KEY1"


# --- completion command test ---


def test_completion_runs():
    """completion command should invoke subprocess."""
    runner = CliRunner()
    with patch("subprocess.run") as mock_run:
        mock_run.return_value.stdout = "# completion script"
        mock_run.return_value.returncode = 0
        result = runner.invoke(main, ["completion", "bash"])
    assert result.exit_code == 0


# --- export --bm25 tests ---


def test_export_bm25_csv(tmp_path: Path):
    from rak.config import RakConfig
    from rak.searcher import SearchResult

    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()

    mock_results = [
        SearchResult(doc_id="BM1", score=5.0, title="BM25 Paper", source="bm25"),
    ]

    with patch("rak.cli.RakConfig", return_value=fake_config), \
         patch("rak.searcher.Searcher.bm25_search", return_value=mock_results), \
         patch("rak.bm25.BM25Index.__init__", return_value=None), \
         patch("rak.bm25.BM25Index.close"):
        result = runner.invoke(main, ["export", "test", "--bm25"])

    assert result.exit_code == 0
    assert "BM1" in result.output
    assert "BM25 Paper" in result.output


# --- config embedding_provider=api display ---


def test_config_show_api_embedding(tmp_path: Path):
    from rak.config import RakConfig, save_config

    save_config(tmp_path, "embedding_provider", "api")
    save_config(tmp_path, "embedding_base_url", "http://example.com/v1")
    save_config(tmp_path, "embedding_api_key", "sk-test12345678")
    fake_config = RakConfig(data_dir=tmp_path)
    runner = CliRunner()
    with patch("rak.cli.RakConfig", return_value=fake_config):
        result = runner.invoke(main, ["config"])

    assert result.exit_code == 0
    assert "embedding_provider = api" in result.output
    assert "embedding_base_url = " in result.output
    assert "embedding_api_key = " in result.output
    # Key should be masked
    assert "sk-test12345678" not in result.output



from __future__ import annotations

import csv
from io import StringIO


def to_csv(results: list[dict]) -> str:
    buf = StringIO()
    writer = csv.writer(buf)
    writer.writerow(["key", "title", "score", "source"])
    for r in results:
        writer.writerow([r["key"], r["title"], round(r["score"], 4), r["source"]])
    return buf.getvalue()


ZOTERO_TO_BIBTEX = {
    "conferencePaper": "inproceedings",
    "bookSection": "incollection",
    "book": "book",
    "thesis": "phdthesis",
    "report": "techreport",
    "presentation": "misc",
    "manuscript": "unpublished",
}


def _bibtex_entry_type(item_type: str) -> str:
    return ZOTERO_TO_BIBTEX.get(item_type, "article")


def _escape_bibtex(value: str) -> str:
    """Escape special BibTeX characters in field values."""
    for char in ('\\', '{', '}', '&', '%', '#', '_', '~', '^'):
        value = value.replace(char, f'\\{char}')
    return value


def _extract_year(date: str) -> str:
    if not date:
        return ""
    return date[:4] if len(date) >= 4 else date


def to_bibtex(results: list[dict]) -> str:
    entries = []
    for r in results:
        entry_type = _bibtex_entry_type(r.get("item_type", ""))
        year = _extract_year(r.get("date", ""))
        title = _escape_bibtex(r['title'])
        authors = _escape_bibtex(r.get('authors', ''))
        entry = (
            f"@{entry_type}{{{r['key']},\n"
            f"  title = {{{title}}},\n"
            f"  author = {{{authors}}},\n"
            f"  year = {{{year}}},\n"
            f"}}"
        )
        entries.append(entry)
    return "\n\n".join(entries)



from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path

DEFAULT_DATA_DIR = Path.home() / "Zotero" / "rak"
DEFAULT_MODEL = "all-MiniLM-L6-v2"


CONFIG_FILENAME = "config.json"
CONFIGURABLE_KEYS = {"llm_base_url", "llm_model", "llm_api_key", "model_name", "zot_command", "chunk_size", "chunk_overlap", "embedding_provider", "embedding_base_url", "embedding_api_key", "embedding_batch_size", "pdf_provider"}
CONFIG_TYPES: dict[str, type] = {"chunk_size": int, "chunk_overlap": int, "embedding_batch_size": int}
VALID_PDF_PROVIDERS = {"pymupdf", "mineru", "docling"}


def _coerce_config_value(key: str, value: str | int | float) -> str | int:
    """Coerce config value to the correct type."""
    target_type = CONFIG_TYPES.get(key)
    if target_type is not None:
        try:
            return target_type(value)
        except (ValueError, TypeError):
            raise ValueError(f"Invalid value for {key}: expected {target_type.__name__}, got {value!r}")
    return str(value)


def _validate_zot_command(value: str) -> None:
    """Reject zot_command values containing path separators or shell metacharacters."""
    import re
    if re.search(r'[/\\;|&$`<>(){}\[\]!#~]', value):
        raise ValueError(f"Invalid zot_command: {value!r} — must be a simple executable name")


def load_config(data_dir: Path) -> dict:
    import json
    path = data_dir / CONFIG_FILENAME
    if not path.exists():
        return {}
    return json.loads(path.read_text())


def _validate_chunk_params(cfg: dict, key: str, value: int) -> None:
    """Validate that chunk_overlap < chunk_size after applying the new value."""
    chunk_size = cfg.get("chunk_size", 512)
    chunk_overlap = cfg.get("chunk_overlap", 64)
    if key == "chunk_size":
        chunk_size = value
    elif key == "chunk_overlap":
        chunk_overlap = value
    if chunk_overlap >= chunk_size:
        raise ValueError(f"chunk_overlap ({chunk_overlap}) must be less than chunk_size ({chunk_size})")
    if chunk_size < 1:
        raise ValueError(f"chunk_size must be positive, got {chunk_size}")


def save_config(data_dir: Path, key: str, value: str) -> None:
    import json
    if key == "zot_command":
        _validate_zot_command(value)
    if key == "pdf_provider" and value not in VALID_PDF_PROVIDERS:
        raise ValueError(f"Invalid pdf_provider: {value!r} — must be one of {sorted(VALID_PDF_PROVIDERS)}")
    coerced = _coerce_config_value(key, value)
    data_dir.mkdir(parents=True, exist_ok=True)
    cfg = load_config(data_dir)
    if key in ("chunk_size", "chunk_overlap"):
        _validate_chunk_params(cfg, key, coerced)
    cfg[key] = coerced
    (data_dir / CONFIG_FILENAME).write_text(json.dumps(cfg, indent=2))


def detect_zotero_storage() -> Path | None:
    storage_dir = Path.home() / "Zotero" / "storage"
    return storage_dir if storage_dir.is_dir() else None


@dataclass
class RakConfig:
    data_dir: Path = DEFAULT_DATA_DIR
    model_name: str = DEFAULT_MODEL
    zot_command: str = "zot"
    zotero_storage_dir: Path | None = field(default_factory=detect_zotero_storage)
    llm_base_url: str = "http://localhost:11434/v1"
    llm_model: str = "qwen3.5:2b"
    llm_api_key: str = "not-needed"
    chunk_size: int = 512
    chunk_overlap: int = 64
    embedding_provider: str = "local"
    embedding_base_url: str = "http://localhost:11434/v1"
    embedding_api_key: str = "not-needed"
    embedding_batch_size: int = 32
    pdf_provider: str = "pymupdf"

    def __post_init__(self) -> None:
        saved = load_config(self.data_dir)
        for key, value in saved.items():
            if key in CONFIGURABLE_KEYS and hasattr(self, key):
                setattr(self, key, _coerce_config_value(key, value))
        # Validate loaded config
        if self.chunk_overlap >= self.chunk_size:
            import logging
            logging.getLogger(__name__).warning(
                "Invalid config: chunk_overlap (%d) >= chunk_size (%d), using defaults",
                self.chunk_overlap, self.chunk_size,
            )
            self.chunk_size = 512
            self.chunk_overlap = 64
        if self.pdf_provider not in VALID_PDF_PROVIDERS:
            import logging
            logging.getLogger(__name__).warning(
                "Invalid config: pdf_provider %r, falling back to 'pymupdf'",
                self.pdf_provider,
            )
            self.pdf_provider = "pymupdf"

    @property
    def chroma_dir(self) -> Path:
        return self.data_dir / "chroma"

    @property
    def fts_db_path(self) -> Path:
        return self.data_dir / "fts.sqlite"



import pytest

from rak.searcher import _deduplicate_chunks, rrf_fuse, SearchResult


def test_rrf_fuse_single_source():
    vector_results = [
        {"id": "A", "score": 0.9},
        {"id": "B", "score": 0.7},
    ]
    fused = rrf_fuse([vector_results], limit=2)
    assert fused[0].doc_id == "A"
    assert fused[1].doc_id == "B"


def test_rrf_fuse_two_sources():
    vector_results = [
        {"id": "A", "score": 0.9},
        {"id": "B", "score": 0.5},
    ]
    bm25_results = [
        {"id": "B", "score": 5.0},
        {"id": "C", "score": 3.0},
    ]
    fused = rrf_fuse([vector_results, bm25_results], limit=3)
    ids = [r.doc_id for r in fused]
    assert "B" in ids[:2]


def test_rrf_fuse_empty():
    fused = rrf_fuse([], limit=5)
    assert fused == []


def test_rrf_fuse_respects_limit():
    results = [{"id": f"X{i}", "score": float(i)} for i in range(20)]
    fused = rrf_fuse([results], limit=5)
    assert len(fused) == 5


def test_search_result_dataclass():
    r = SearchResult(doc_id="A1", score=0.85, title="Test Paper", source="vector")
    assert r.doc_id == "A1"
    assert r.score == 0.85


from rak.searcher import build_where_filter


def test_build_where_filter_collection_only():
    f = build_where_filter(collection="My Papers", tags=None)
    assert f == {"collections": {"$contains": "My Papers"}}


def test_build_where_filter_single_tag():
    f = build_where_filter(collection=None, tags=["RNA"])
    assert f == {"tags": {"$contains": "RNA"}}


def test_build_where_filter_multiple_tags():
    f = build_where_filter(collection=None, tags=["RNA", "DNA"])
    assert f == {"$or": [{"tags": {"$contains": "RNA"}}, {"tags": {"$contains": "DNA"}}]}


def test_build_where_filter_collection_and_tags():
    f = build_where_filter(collection="Bio", tags=["RNA"])
    assert f == {"$and": [{"collections": {"$contains": "Bio"}}, {"tags": {"$contains": "RNA"}}]}


def test_build_where_filter_none():
    f = build_where_filter(collection=None, tags=None)
    assert f is None


def test_deduplicate_chunks_merges_same_parent():
    results = [
        SearchResult(doc_id="ABC_chunk_0", score=0.9, title="Paper A", source="vector", snippet="best chunk"),
        SearchResult(doc_id="ABC_chunk_1", score=0.8, title="Paper A", source="vector", snippet="other chunk"),
        SearchResult(doc_id="DEF", score=0.7, title="Paper B", source="vector", snippet="only chunk"),
    ]
    deduped = _deduplicate_chunks(results)
    ids = [r.doc_id for r in deduped]
    assert ids == ["ABC", "DEF"]
    assert deduped[0].score == 0.9
    assert deduped[0].snippet == "best chunk"


def test_rrf_fuse_preserves_snippets():
    vector_results = [
        {"id": "A", "score": 0.9, "document": "snippet text A"},
        {"id": "B", "score": 0.7, "document": "snippet text B"},
    ]
    bm25_results = [
        {"id": "B", "score": 5.0},
        {"id": "C", "score": 3.0},
    ]
    fused = rrf_fuse([vector_results, bm25_results], limit=3)
    snippets = {r.doc_id: r.snippet for r in fused}
    assert snippets["A"] == "snippet text A"
    assert snippets["B"] == "snippet text B"
    assert snippets["C"] == ""


def test_bm25_search_standalone():
    """Test BM25-only search via Searcher (no embedder/vector store needed)."""
    from unittest.mock import MagicMock
    bm25 = MagicMock()
    bm25.search_with_snippet.return_value = [
        {"id": "A1", "score": 5.0, "snippet": "RNA analysis...", "title": ""},
        {"id": "B1", "score": 3.0, "snippet": "cell method...", "title": ""},
    ]
    from rak.searcher import Searcher
    searcher = Searcher(None, None, bm25)
    results = searcher.bm25_search("RNA cell", limit=10)
    assert len(results) == 2
    assert results[0].doc_id == "A1"
    assert results[0].source == "bm25"
    assert results[0].snippet == "RNA analysis..."


def test_bm25_search_deduplicates_chunks():
    from unittest.mock import MagicMock
    bm25 = MagicMock()
    bm25.search_with_snippet.return_value = [
        {"id": "A1_chunk_0", "score": 5.0, "snippet": "best chunk", "title": ""},
        {"id": "A1_chunk_1", "score": 3.0, "snippet": "other chunk", "title": ""},
        {"id": "B1", "score": 2.0, "snippet": "another paper", "title": ""},
    ]
    from rak.searcher import Searcher
    searcher = Searcher(None, None, bm25)
    results = searcher.bm25_search("test", limit=10)
    ids = [r.doc_id for r in results]
    assert ids == ["A1", "B1"]


def test_deduplicate_chunks_no_chunks():
    results = [
        SearchResult(doc_id="A", score=0.9, title="Paper", source="vector"),
        SearchResult(doc_id="B", score=0.8, title="Paper 2", source="vector"),
    ]
    deduped = _deduplicate_chunks(results)
    assert len(deduped) == 2
    assert deduped[0].doc_id == "A"


def test_similar_search_excludes_source():
    """similar_search returns similar papers excluding the source paper."""
    from unittest.mock import MagicMock
    from rak.searcher import Searcher

    vector_store = MagicMock()
    vector_store.get_embedding.return_value = [0.1, 0.2, 0.3]
    vector_store.search.return_value = [
        {"id": "SRC", "score": 1.0, "document": "source paper", "metadata": {"title": "Source"}},
        {"id": "A", "score": 0.9, "document": "similar paper A", "metadata": {"title": "Paper A"}},
        {"id": "B", "score": 0.8, "document": "similar paper B", "metadata": {"title": "Paper B"}},
    ]
    bm25 = MagicMock()
    searcher = Searcher(MagicMock(), vector_store, bm25)
    results = searcher.similar_search("SRC", limit=5)
    ids = [r.doc_id for r in results]
    assert "SRC" not in ids
    assert "A" in ids
    assert "B" in ids
    assert results[0].source == "similar"


def test_similar_search_uses_chunk_embedding():
    """similar_search falls back to chunk_0 embedding when key not found."""
    from unittest.mock import MagicMock
    from rak.searcher import Searcher

    vector_store = MagicMock()
    vector_store.get_embedding.side_effect = lambda key: None if key == "SRC" else [0.1, 0.2]
    vector_store.search.return_value = [
        {"id": "SRC_chunk_0", "score": 1.0, "document": "chunk", "metadata": {"title": "Source"}},
        {"id": "A", "score": 0.9, "document": "other", "metadata": {"title": "Paper A"}},
    ]
    bm25 = MagicMock()
    searcher = Searcher(MagicMock(), vector_store, bm25)
    results = searcher.similar_search("SRC", limit=5)
    vector_store.get_embedding.assert_any_call("SRC_chunk_0")
    ids = [r.doc_id for r in results]
    assert "SRC" not in ids


def test_similar_search_key_not_found():
    """similar_search returns empty list when key doesn't exist."""
    from unittest.mock import MagicMock
    from rak.searcher import Searcher

    vector_store = MagicMock()
    vector_store.get_embedding.return_value = None
    bm25 = MagicMock()
    searcher = Searcher(MagicMock(), vector_store, bm25)
    results = searcher.similar_search("NONEXIST", limit=5)
    assert results == []


def test_similar_search_no_vector_store():
    """similar_search returns empty when no vector store (BM25-only mode)."""
    from unittest.mock import MagicMock
    from rak.searcher import Searcher

    bm25 = MagicMock()
    searcher = Searcher(None, None, bm25)
    results = searcher.similar_search("SRC", limit=5)
    assert results == []


def test_similar_search_with_filters():
    """similar_search passes where filter to vector_store.search when collection/tags given."""
    from unittest.mock import MagicMock, call
    from rak.searcher import Searcher, build_where_filter

    vector_store = MagicMock()
    vector_store.get_embedding.return_value = [0.1, 0.2, 0.3]
    vector_store.search.return_value = [
        {"id": "A", "score": 0.9, "document": "paper A", "metadata": {"title": "Paper A"}},
    ]
    bm25 = MagicMock()
    searcher = Searcher(MagicMock(), vector_store, bm25)

    results = searcher.similar_search("SRC", limit=5, collection="My Collection", tags=["RNA"])

    expected_where = build_where_filter(collection="My Collection", tags=["RNA"])
    vector_store.search.assert_called_once_with(
        [0.1, 0.2, 0.3], limit=18, where=expected_where
    )
    assert results[0].doc_id == "A"



from __future__ import annotations

from dataclasses import dataclass

from rak.bm25 import BM25Index
from rak.embedder import Embedder
from rak.store import VectorStore


@dataclass
class SearchResult:
    doc_id: str
    score: float
    title: str = ""
    source: str = ""
    snippet: str = ""


def rrf_fuse(
    ranked_lists: list[list[dict]],
    limit: int = 10,
    k: int = 60,  # Standard RRF constant; higher k reduces impact of top ranks (60 is the value from the original Cormack et al. paper)
) -> list[SearchResult]:
    scores: dict[str, float] = {}
    titles: dict[str, str] = {}
    snippets: dict[str, str] = {}
    for ranked_list in ranked_lists:
        for rank, item in enumerate(ranked_list):
            doc_id = item["id"]
            scores[doc_id] = scores.get(doc_id, 0.0) + 1.0 / (k + rank + 1)
            if "metadata" in item and "title" in item["metadata"]:
                titles[doc_id] = item["metadata"]["title"]
            elif "title" in item:
                titles[doc_id] = item["title"]
            if doc_id not in snippets and item.get("document"):
                snippets[doc_id] = item["document"]
    sorted_ids = sorted(scores, key=lambda x: scores[x], reverse=True)[:limit]
    return [
        SearchResult(
            doc_id=doc_id, score=scores[doc_id], title=titles.get(doc_id, ""),
            source="fused", snippet=snippets.get(doc_id, ""),
        )
        for doc_id in sorted_ids
    ]


def build_where_filter(
    collection: str | None = None,
    tags: list[str] | None = None,
) -> dict | None:
    filters = []
    if collection:
        filters.append({"collections": {"$contains": collection}})
    if tags:
        if len(tags) == 1:
            filters.append({"tags": {"$contains": tags[0]}})
        else:
            filters.append({"$or": [{"tags": {"$contains": t}} for t in tags]})
    if not filters:
        return None
    if len(filters) == 1:
        return filters[0]
    return {"$and": filters}


def _deduplicate_chunks(results: list[SearchResult]) -> list[SearchResult]:
    """Deduplicate chunk results back to parent papers, keeping the best score and snippet."""
    seen: dict[str, SearchResult] = {}
    for r in results:
        parent_key = r.doc_id.split("_chunk_")[0] if "_chunk_" in r.doc_id else r.doc_id
        if parent_key not in seen or r.score > seen[parent_key].score:
            seen[parent_key] = SearchResult(
                doc_id=parent_key, score=r.score, title=r.title, source=r.source,
                snippet=r.snippet,
            )
    return sorted(seen.values(), key=lambda x: x.score, reverse=True)


class Searcher:
    def __init__(self, embedder: Embedder | None, vector_store: VectorStore | None, bm25_index: BM25Index) -> None:
        self._embedder = embedder
        self._vector_store = vector_store
        self._bm25 = bm25_index

    def bm25_search(
        self, query: str, limit: int = 10,
    ) -> list[SearchResult]:
        """Pure BM25 keyword search. Does not require embedder or vector store."""
        results = self._bm25.search_with_snippet(query, limit=limit)
        raw = [
            SearchResult(
                doc_id=r["id"], score=r["score"],
                title=r.get("title", ""), source="bm25",
                snippet=r.get("snippet", ""),
            )
            for r in results
        ]
        return _deduplicate_chunks(raw)[:limit]

    def vector_search(
        self, query: str, limit: int = 10,
        collection: str | None = None, tags: list[str] | None = None,
    ) -> list[SearchResult]:
        embedding = self._embedder.embed(query)
        where = build_where_filter(collection, tags)
        # Fetch 3x results to ensure enough unique papers after chunk deduplication
        # (multiple chunks per paper are stored as separate vectors)
        results = self._vector_store.search(embedding, limit=limit * 3, where=where)
        raw = [
            SearchResult(
                doc_id=r["id"], score=r["score"],
                title=r.get("metadata", {}).get("title", ""),
                source="vector",
                snippet=r.get("document", ""),
            )
            for r in results
        ]
        return _deduplicate_chunks(raw)[:limit]

    def similar_search(
        self, doc_key: str, limit: int = 10,
        collection: str | None = None, tags: list[str] | None = None,
    ) -> list[SearchResult]:
        """Find papers similar to a given document by its key."""
        if self._vector_store is None:
            return []
        # Try the key directly, then try the first chunk
        embedding = self._vector_store.get_embedding(doc_key)
        if embedding is None:
            embedding = self._vector_store.get_embedding(f"{doc_key}_chunk_0")
        if embedding is None:
            return []
        where = build_where_filter(collection, tags)
        # Fetch extra to account for chunks and self-exclusion
        results = self._vector_store.search(embedding, limit=(limit + 1) * 3, where=where)
        raw = [
            SearchResult(
                doc_id=r["id"], score=r["score"],
                title=r.get("metadata", {}).get("title", ""),
                source="similar",
                snippet=r.get("document", ""),
            )
            for r in results
        ]
        deduped = _deduplicate_chunks(raw)
        # Exclude the source paper
        return [r for r in deduped if r.doc_id != doc_key][:limit]

    def hybrid_search(
        self, query: str, limit: int = 10,
        collection: str | None = None, tags: list[str] | None = None,
    ) -> list[SearchResult]:
        embedding = self._embedder.embed(query)
        where = build_where_filter(collection, tags)
        vector_results = self._vector_store.search(embedding, limit=limit * 3, where=where)
        bm25_results = self._bm25.search(query, limit=limit * 2)
        fused = rrf_fuse([vector_results, bm25_results], limit=limit * 3)
        return _deduplicate_chunks(fused)[:limit]



from rak.export import to_csv, to_bibtex


def test_to_csv():
    results = [
        {"key": "A1", "title": "Paper One", "score": 0.85, "source": "vector", "date": "2024", "authors": "Doe, J."},
        {"key": "A2", "title": "Paper Two", "score": 0.72, "source": "fused", "date": "2023", "authors": "Smith, A."},
    ]
    csv_text = to_csv(results)
    lines = [l.strip() for l in csv_text.strip().splitlines()]
    assert lines[0] == "key,title,score,source"
    assert "A1" in lines[1]
    assert "Paper One" in lines[1]
    assert "0.85" in lines[1]


def test_to_csv_empty():
    csv_text = to_csv([])
    lines = [l.strip() for l in csv_text.strip().splitlines()]
    assert lines[0] == "key,title,score,source"
    assert len(lines) == 1


def test_to_bibtex():
    results = [
        {"key": "A1", "title": "Paper One", "score": 0.85, "source": "vector", "date": "2024", "authors": "Doe, J."},
    ]
    bib = to_bibtex(results)
    assert "@article{A1," in bib
    assert "title = {Paper One}" in bib
    assert "year = {2024}" in bib
    assert "author = {Doe, J.}" in bib


def test_to_bibtex_empty():
    bib = to_bibtex([])
    assert bib == ""


def test_to_bibtex_conference_paper():
    results = [
        {"key": "C1", "title": "Conf Paper", "score": 0.9, "source": "vector",
         "date": "2024-06-15", "authors": "Lee, K.", "item_type": "conferencePaper"},
    ]
    bib = to_bibtex(results)
    assert "@inproceedings{C1," in bib
    assert "year = {2024}" in bib


def test_to_bibtex_book():
    results = [
        {"key": "B1", "title": "A Book", "score": 0.7, "source": "vector",
         "date": "2020", "authors": "Author", "item_type": "book"},
    ]
    bib = to_bibtex(results)
    assert "@book{B1," in bib


def test_to_bibtex_unknown_type_defaults_to_article():
    results = [
        {"key": "U1", "title": "Unknown", "score": 0.5, "source": "vector",
         "date": "2023", "authors": "X", "item_type": "weirdType"},
    ]
    bib = to_bibtex(results)
    assert "@article{U1," in bib
